from collections.abc import ItemsView, Iterable, Iterator, KeysView, Mapping, Sequence, ValuesView
from enum import Enum
from typing import (
    Callable,
    Optional,
    Union,
)


class FeatureType(Enum):
    """Type of Feature."""

    CATEGORICAL = "categorical"
    CATEGORICAL_LIST = "categorical_list"
    NUMERICAL = "numerical"
    NUMERICAL_LIST = "numerical_list"


class FeatureSource(Enum):
    """Name of DataFrame."""

    ITEM_FEATURES = "item_features"
    QUERY_FEATURES = "query_features"
    INTERACTIONS = "interactions"


class FeatureHint(Enum):
    """Hint to algorithm about column."""

    ITEM_ID = "item_id"
    QUERY_ID = "query_id"
    RATING = "rating"
    TIMESTAMP = "timestamp"


class FeatureInfo:
    """
    Information about a feature.
    """

    def __init__(
        self,
        column: str,
        feature_type: FeatureType,
        feature_hint: Optional[FeatureHint] = None,
        feature_source: Optional[FeatureSource] = None,
        cardinality: Optional[int] = None,
    ) -> None:
        """
        :param column: name of feature.
        :param feature_type: type of feature.
        :param feature_hint: hint to models about feature
            (is timestamp, is rating, is query_id, is item_id),
            default: ``None``.
        :param feature_source: name of DataFrame feature came from,
            default: ``None``.
        :param cardinality: cardinality of categorical feature, required for ids columns,
            optional for others,
            default: ``None``.
        """
        self._column = column
        self._feature_type = feature_type
        self._feature_source = feature_source
        self._feature_hint = feature_hint

        if feature_type in [FeatureType.NUMERICAL, FeatureType.NUMERICAL_LIST] and cardinality:
            msg = "Cardinality is needed only with categorical feature_type."
            raise ValueError(msg)
        self._cardinality = cardinality

    @property
    def column(self) -> str:
        """
        :returns: the feature name.
        """
        return self._column

    @property
    def feature_type(self) -> FeatureType:
        """
        :returns: the type of feature.
        """
        return self._feature_type

    @property
    def feature_hint(self) -> Optional[FeatureHint]:
        """
        :returns: the feature hint.
        """
        return self._feature_hint

    @property
    def feature_source(self) -> Optional[FeatureSource]:
        """
        :returns: the name of source dataframe of feature.
        """
        return self._feature_source

    def _set_feature_source(self, source: FeatureSource) -> None:
        self._feature_source = source

    @property
    def cardinality(self) -> Optional[int]:
        """
        :returns: cardinality of the feature.
        """
        if self.feature_type not in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST]:
            msg = f"Can not get cardinality because feature_type of {self.column} column is not categorical."
            raise RuntimeError(msg)
        if hasattr(self, "_cardinality_callback") and self._cardinality is None:
            self._cardinality = self._cardinality_callback(self._column)
        return self._cardinality

    def _set_cardinality_callback(self, callback: Callable) -> None:
        self._cardinality_callback = callback

    def reset_cardinality(self) -> None:
        """
        Reset cardinality of the feature to None.
        """
        self._cardinality = None


class FeatureSchema(Mapping[str, FeatureInfo]):
    """
    Key-value like collection with information about all dataset features.
    """

    def __init__(self, features_list: Union[Sequence[FeatureInfo], FeatureInfo]) -> None:
        """
        :param features_list: list of feature infos.
        """
        features_list = [features_list] if not isinstance(features_list, Sequence) else features_list
        self._check_features_naming(features_list)
        self._features_schema = {feature.column: feature for feature in features_list}

    def copy(self) -> "FeatureSchema":
        """
        Creates a copy of all features. For the returned copy, all cardinality values will be undefined.

        :returns: copy of the initial feature schema.
        """
        copy_features_list = list(self._features_schema.values())
        for feature in copy_features_list:
            feature.reset_cardinality()
        return FeatureSchema(copy_features_list)

    def subset(self, features_to_keep: Iterable[str]) -> "FeatureSchema":
        """
        Creates a subset of given features.

        :param features_to_keep: a sequence of feature columns
            in original schema to keep in subset.
        :returns: new feature schema of given features.
        """
        features: set[FeatureInfo] = set()
        for feature_column in features_to_keep:
            if feature_column in self._features_schema:
                features.add(self._features_schema[feature_column])
        return FeatureSchema(list(features))

    def item(self) -> FeatureInfo:
        """
        :returns: extract a feature information from a schema.
        """
        if len(self._features_schema) > 1:
            msg = "Only one element feature schema can be converted to single feature"
            raise ValueError(msg)
        return next(iter(self._features_schema.values()))

    def items(self) -> ItemsView[str, FeatureInfo]:
        return self._features_schema.items()

    def keys(self) -> KeysView[str]:
        return self._features_schema.keys()

    def values(self) -> ValuesView[FeatureInfo]:
        return self._features_schema.values()

    def get(
        self,
        key: str,
        default: Optional[FeatureInfo] = None,
    ) -> Optional[FeatureInfo]:
        return self._features_schema.get(key, default)

    def __iter__(self) -> Iterator[str]:
        return iter(self._features_schema)

    def __contains__(self, feature_name: object) -> bool:
        return feature_name in self._features_schema

    def __len__(self) -> int:
        return len(self._features_schema)

    def __bool__(self) -> bool:
        return len(self._features_schema) > 0

    def __getitem__(self, feature_name: str) -> FeatureInfo:
        return self._features_schema[feature_name]

    def __eq__(self, other: object) -> bool:
        return self._features_schema == other

    def __ne__(self, other: object) -> bool:
        return self._features_schema != other

    def __add__(self, other: "FeatureSchema") -> "FeatureSchema":
        return FeatureSchema(list(self._features_schema.values()) + list(other._features_schema.values()))

    @property
    def all_features(self) -> Sequence[FeatureInfo]:
        """
        :returns: sequence of all features.
        """
        return list(self._features_schema.values())

    @property
    def categorical_features(self) -> "FeatureSchema":
        """
        :returns: sequence of categorical features in a schema.
        """
        return self.filter(feature_type=FeatureType.CATEGORICAL) + self.filter(
            feature_type=FeatureType.CATEGORICAL_LIST
        )

    @property
    def numerical_features(self) -> "FeatureSchema":
        """
        :returns: sequence of numerical features in a schema.
        """
        return self.filter(feature_type=FeatureType.NUMERICAL) + self.filter(feature_type=FeatureType.NUMERICAL_LIST)

    @property
    def interaction_features(self) -> "FeatureSchema":
        """
        :returns: sequence of interaction features in a schema.
        """
        return (
            self.filter(feature_source=FeatureSource.INTERACTIONS)
            .drop(feature_hint=FeatureHint.ITEM_ID)
            .drop(feature_hint=FeatureHint.QUERY_ID)
        )

    @property
    def query_features(self) -> "FeatureSchema":
        """
        :returns: sequence of query features in a schema.
        """
        return self.filter(feature_source=FeatureSource.QUERY_FEATURES)

    @property
    def item_features(self) -> "FeatureSchema":
        """
        :returns: sequence of item features in a schema.
        """
        return self.filter(feature_source=FeatureSource.ITEM_FEATURES)

    @property
    def interactions_rating_features(self) -> "FeatureSchema":
        """
        :returns: sequence of interactions-rating features in a schema.
        """
        return self.filter(feature_source=FeatureSource.INTERACTIONS, feature_hint=FeatureHint.RATING)

    @property
    def interactions_timestamp_features(self) -> "FeatureSchema":
        """
        :returns: sequence of interactions-timestamp features in a schema.
        """
        return self.filter(feature_source=FeatureSource.INTERACTIONS, feature_hint=FeatureHint.TIMESTAMP)

    @property
    def columns(self) -> Sequence[str]:
        """
        :returns: list of all feature's column names.
        """
        return list(self._features_schema)

    @property
    def query_id_feature(self) -> FeatureInfo:
        """
        :returns: sequence of query id features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.QUERY_ID).item()

    @property
    def item_id_feature(self) -> FeatureInfo:
        """
        :returns: sequence of item id features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.ITEM_ID).item()

    @property
    def query_id_column(self) -> str:
        """
        :returns: query id column name.
        """
        return self.query_id_feature.column

    @property
    def item_id_column(self) -> str:
        """
        :returns: item id column name.
        """
        return self.item_id_feature.column

    @property
    def interactions_rating_column(self) -> Optional[str]:
        """
        :returns: interactions-rating column name.
        """
        interactions_rating_features = self.interactions_rating_features
        if not interactions_rating_features:
            return None
        return interactions_rating_features.item().column

    @property
    def interactions_timestamp_column(self) -> Optional[str]:
        """
        :returns: interactions-timestamp column name.
        """
        interactions_timestamp_features = self.interactions_timestamp_features
        if not interactions_timestamp_features:
            return None
        return interactions_timestamp_features.item().column

    def filter(
        self,
        column: Optional[str] = None,
        feature_hint: Optional[FeatureHint] = None,
        feature_source: Optional[FeatureSource] = None,
        feature_type: Optional[FeatureType] = None,
    ) -> "FeatureSchema":
        """Filter list by ``column``, ``feature_source``, ``feature_type`` and ``feature_hint``.

        :param column: Column name to filter by.
            default: ``None``.
        :param feature_hint: Feature hint to filter by.
            default: ``None``.
        :param feature_source: Feature source to filter by.
            default: ``None``.
        :param feature_type: Feature type to filter by.
            default: ``None``.

        :returns: new filtered feature schema.
        """
        filtered_features = self.all_features
        filter_functions = [self._name_filter, self._source_filter, self._type_filter, self._hint_filter]
        filter_parameters = [column, feature_source, feature_type, feature_hint]
        for filtration_func, filtration_param in zip(filter_functions, filter_parameters):
            filtered_features = list(
                filter(
                    lambda x: filtration_func(x, filtration_param),
                    filtered_features,
                )
            )

        return FeatureSchema(filtered_features)

    def drop(
        self,
        column: Optional[str] = None,
        feature_hint: Optional[FeatureHint] = None,
        feature_source: Optional[FeatureSource] = None,
        feature_type: Optional[FeatureType] = None,
    ) -> "FeatureSchema":
        """Drop features from list by ``column``, ``feature_source``, ``feature_type`` and ``feature_hint``.

        :param column: Column name to filter by.
            default: ``None``.
        :param feature_hint: Feature hint to filter by.
            default: ``None``.
        :param feature_source: Feature source to filter by.
            default: ``None``.
        :param feature_type: Feature type to filter by.
            default: ``None``.

        :returns: new filtered feature schema without selected features.
        """
        filtered_features = self.all_features
        filter_functions = [self._name_drop, self._source_drop, self._type_drop, self._hint_drop]
        filter_parameters = [column, feature_source, feature_type, feature_hint]
        for filtration_func, filtration_param in zip(filter_functions, filter_parameters):
            filtered_features = list(
                filter(
                    lambda x: filtration_func(x, filtration_param),
                    filtered_features,
                )
            )

        return FeatureSchema(filtered_features)

    @staticmethod
    def _name_filter(value: FeatureInfo, column: str) -> bool:
        return value.column == column if column else True

    @staticmethod
    def _source_filter(value: FeatureInfo, feature_source: FeatureSource) -> bool:
        return value.feature_source == feature_source if feature_source else True

    @staticmethod
    def _type_filter(value: FeatureInfo, feature_type: FeatureType) -> bool:
        return value.feature_type == feature_type if feature_type else True

    @staticmethod
    def _hint_filter(value: FeatureInfo, feature_hint: FeatureHint) -> bool:
        return value.feature_hint == feature_hint if feature_hint else True

    @staticmethod
    def _name_drop(value: FeatureInfo, column: str) -> bool:
        return value.column != column if column else True

    @staticmethod
    def _source_drop(value: FeatureInfo, feature_source: FeatureSource) -> bool:
        return value.feature_source != feature_source if feature_source else True

    @staticmethod
    def _type_drop(value: FeatureInfo, feature_type: FeatureType) -> bool:
        return value.feature_type != feature_type if feature_type else True

    @staticmethod
    def _hint_drop(value: FeatureInfo, feature_hint: FeatureHint) -> bool:
        return value.feature_hint != feature_hint if feature_hint else True

    def _check_features_naming(self, features_list: Sequence[FeatureInfo]) -> None:
        """
        Checks that all the columns have unique names except QUERY_ID and ITEM_ID columns.
        """
        unique_columns = set()
        duplicates = set()
        item_query_names: dict[FeatureHint, list[str]] = {
            FeatureHint.ITEM_ID: [],
            FeatureHint.QUERY_ID: [],
        }
        for feature in features_list:
            if feature.feature_hint not in [FeatureHint.ITEM_ID, FeatureHint.QUERY_ID]:
                if feature.column in unique_columns:
                    duplicates.add(feature.column)
                else:
                    unique_columns.add(feature.column)
            else:
                item_query_names[feature.feature_hint] += [feature.column]

        if len(duplicates) > 0:
            msg = (
                "Features column names should be unique, except ITEM_ID and QUERY_ID columns. "
                f"{duplicates} columns are not unique."
            )
            raise ValueError(msg)

        if len(item_query_names[FeatureHint.ITEM_ID]) > 1:
            msg = f"ITEM_ID must be present only once. Rename {item_query_names[FeatureHint.ITEM_ID]}"
            raise ValueError(msg)

        if len(item_query_names[FeatureHint.QUERY_ID]) > 1:
            msg = f"QUERY_ID must be present only once. Rename {item_query_names[FeatureHint.QUERY_ID]}"
            raise ValueError(msg)



from abc import ABC, abstractmethod
from typing import Any, Callable


class IndexStore(ABC):
    """Abstract base class for index stores. Describes a common interface for index stores."""

    def __init__(self, cleanup: bool = True):
        self._index = None
        self.cleanup = cleanup

    @abstractmethod
    def save_to_store(self, save_index: Callable[[str], None]):
        """Dumps index file to store"""

    @abstractmethod
    def load_index(
        self,
        init_index: Callable[[], None],
        load_index: Callable[[Any, str], None],
        configure_index: Callable[[Any], None],
    ) -> Any:
        """Loads index from IndexStore to index instance.
        If the index has already been loaded before,
        the previously loaded index instance is returned."""

    @abstractmethod
    def dump_index(self, target_path: str):
        """Dumps index files to `target_path`"""



from typing import Optional

from scipy.stats import norm

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE

from .pop_rec import PopRec

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class Wilson(PopRec):
    """
    Calculates lower confidence bound for the confidence interval
    of true fraction of positive ratings.

    ``rating`` must be converted to binary 0-1 form.

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame({"user_id": [1, 2], "item_id": [1, 2], "rating": [1, 1]})
    >>> interactions = convert2spark(data_frame)
    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> dataset = Dataset(feature_schema, interactions)
    >>> model = Wilson()
    >>> model.fit_predict(dataset, k=1).toPandas()
        user_id   item_id     rating
    0         1         2   0.206549
    1         2         1   0.206549

    """

    def __init__(
        self,
        alpha=0.05,
        add_cold_items: bool = True,
        cold_weight: float = 0.5,
        sample: bool = False,
        seed: Optional[int] = None,
    ):
        """
        :param alpha: significance level, default 0.05
        :param add_cold_items: flag to consider cold items in recommendations building
            if present in `items` parameter of `predict` method
            or `pairs` parameter of `predict_pairs` methods.
            If true, cold items are assigned rating equals to the less relevant item rating
            multiplied by cold_weight and may appear among top-K recommendations.
            Otherwise cold items are filtered out.
            Could be changed after model training by setting the `add_cold_items` attribute.
        : param cold_weight: if `add_cold_items` is True,
            cold items are added with reduced rating.
            The rating for cold items is equal to the rating
            of a least relevant item multiplied by a `cold_weight` value.
            `Cold_weight` value should be in interval (0, 1].
        :param sample: flag to choose recommendation strategy.
            If True, items are sampled with a probability proportional
            to the calculated predicted rating.
            Could be changed after model training by setting the `sample` attribute.
        :param seed: random seed. Provides reproducibility if fixed
        """
        self.alpha = alpha
        self.sample = sample
        self.seed = seed
        super().__init__(add_cold_items=add_cold_items, cold_weight=cold_weight)

    @property
    def _init_args(self):
        return {
            "alpha": self.alpha,
            "add_cold_items": self.add_cold_items,
            "cold_weight": self.cold_weight,
            "sample": self.sample,
            "seed": self.seed,
        }

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self._check_rating(dataset)

        items_counts = dataset.interactions.groupby(self.item_column).agg(
            sf.sum(self.rating_column).alias("pos"),
            sf.count(self.rating_column).alias("total"),
        )
        # https://en.wikipedia.org/w/index.php?title=Binomial_proportion_confidence_interval
        crit = norm.isf(self.alpha / 2.0)
        items_counts = items_counts.withColumn(
            self.rating_column,
            (sf.col("pos") + sf.lit(0.5 * crit**2)) / (sf.col("total") + sf.lit(crit**2))
            - sf.lit(crit)
            / (sf.col("total") + sf.lit(crit**2))
            * sf.sqrt((sf.col("total") - sf.col("pos")) * sf.col("pos") / sf.col("total") + crit**2 / 4),
        )

        self.item_popularity = items_counts.drop("pos", "total")
        self.item_popularity.cache().count()
        self.fill = self._calc_fill(self.item_popularity, self.cold_weight, self.rating_column)



from .types import PYSPARK_AVAILABLE, SparkDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.ml import Transformer
    from pyspark.ml.param import Param, Params, TypeConverters
    from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

    from replay.utils.model_handler import get_fs
    from replay.utils.session_handler import State


class DataframeBucketizer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    """
    Buckets the input dataframe, dumps it to spark warehouse directory,
    and returns a bucketed dataframe.
    """

    bucketing_key = Param(
        Params._dummy(),
        "bucketing_key",
        "bucketing key (also used as sort key)",
        typeConverter=TypeConverters.toString,
    )

    partition_num = Param(
        Params._dummy(),
        "partition_num",
        "number of buckets",
        typeConverter=TypeConverters.toInt,
    )

    table_name = Param(
        Params._dummy(),
        "table_name",
        "parquet file name (for storage  in 'spark-warehouse') and spark table name",
        typeConverter=TypeConverters.toString,
    )

    spark_warehouse_dir = Param(
        Params._dummy(),
        "spark_warehouse_dir",
        "spark_warehouse_dir",
        typeConverter=TypeConverters.toString,
    )

    def __init__(
        self,
        bucketing_key: str,
        partition_num: int,
        spark_warehouse_dir: str,
        table_name: str = "",
    ):
        """Makes bucketed dataframe from input dataframe.

        Args:
            bucketing_key: bucketing key (also used as sort key)
            partition_num: number of buckets
            table_name: parquet file name (for storage  in 'spark-warehouse') and spark table name
            spark_warehouse_dir: spark warehouse dir,
                i.e. value of 'spark.sql.warehouse.dir' property
        """
        super().__init__()
        self.set(self.bucketing_key, bucketing_key)
        self.set(self.partition_num, partition_num)
        self.set(self.table_name, table_name)
        self.set(self.spark_warehouse_dir, spark_warehouse_dir)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.remove_parquet()

    def remove_parquet(self):
        """Removes parquets where bucketed dataset is stored"""
        spark = State().session
        spark_warehouse_dir = self.getOrDefault(self.spark_warehouse_dir)
        table_name = self.getOrDefault(self.table_name)
        fs = get_fs(spark)
        fs_path = spark._jvm.org.apache.hadoop.fs.Path(f"{spark_warehouse_dir}/{table_name}")
        is_exists = fs.exists(fs_path)
        if is_exists:
            fs.delete(fs_path, True)

    def set_table_name(self, table_name: str):
        """Sets table name"""
        self.set(self.table_name, table_name)

    def _transform(self, dataset: SparkDataFrame):
        bucketing_key = self.getOrDefault(self.bucketing_key)
        partition_num = self.getOrDefault(self.partition_num)
        table_name = self.getOrDefault(self.table_name)
        spark_warehouse_dir = self.getOrDefault(self.spark_warehouse_dir)

        if not table_name:
            msg = "Parameter 'table_name' is not set! Please set it via method 'set_table_name'."
            raise ValueError(msg)

        (
            dataset.repartition(partition_num, bucketing_key)
            .write.mode("overwrite")
            .bucketBy(partition_num, bucketing_key)
            .sortBy(bucketing_key)
            .saveAsTable(
                table_name,
                format="parquet",
                path=f"{spark_warehouse_dir}/{table_name}",
            )
        )

        spark = State().session

        return spark.table(table_name)



import pathlib
import tempfile
from abc import abstractmethod
from typing import Any, Literal, Optional, Union

import lightning
import openvino as ov
import torch

from replay.data.nn import TensorSchema

OptimizedModeType = Literal[
    "batch",
    "one_query",
    "dynamic_batch_size",
]


def _compile_openvino(
    onnx_path: str,
    batch_size: int,
    max_seq_len: int,
    num_candidates_to_score: int,
    num_threads: Optional[int],
) -> ov.CompiledModel:
    """
    Method defines compilation strategy for openvino backend.

    :param onnx_path: Path to the model representation in ONNX format.
    :param batch_size: Defines whether batch will be static or dynamic length.
    :param max_seq_len: Defines whether sequence will be static or dynamic length.
    :param num_candidates_to_score: Defines whether candidates will be static or dynamic length.
    :param num_threads: Defines number of CPU threads for which the model will be compiled by the OpenVino core.
        If ``None``, then compiler will set this parameter automatically.
        Default: ``None``.

    :return: Compiled model.
    """
    core = ov.Core()
    if num_threads is not None:
        core.set_property("CPU", {"INFERENCE_NUM_THREADS": num_threads})
    model_onnx = core.read_model(model=onnx_path)
    inputs_names = [inputs.names.pop() for inputs in model_onnx.inputs]
    del model_onnx

    candidates_input_id = len(inputs_names) - 1 if num_candidates_to_score is not None else len(inputs_names)
    model_input_scheme = [(input_name, [batch_size, max_seq_len]) for input_name in inputs_names[:candidates_input_id]]
    if num_candidates_to_score is not None:
        model_input_scheme += [(inputs_names[candidates_input_id], [num_candidates_to_score])]
    model_onnx = ov.convert_model(onnx_path, input=model_input_scheme)
    return core.compile_model(model=model_onnx, device_name="CPU")


class BaseCompiledModel:
    """
    Base class of CPU-optimized model for inference via OpenVINO.
    It is recommended to use inherited classes and not to use this one.
    """

    def __init__(
        self,
        compiled_model: ov.CompiledModel,
        schema: TensorSchema,
    ) -> None:
        """
        :param compiled_model: Compiled model.
        :param schema: Tensor schema of SasRec model.
        """
        self._batch_size: int
        self._max_seq_len: int
        self._inputs_names: list[str]
        self._output_name: str

        self._set_inner_params_from_openvino_model(compiled_model)
        self._schema = schema
        self._model = compiled_model

    @abstractmethod
    def predict(
        self,
        batch: Any,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        Inference on one batch.

        :param batch: Prediction input.
        :param candidates_to_score: Item ids to calculate scores.
            Default: ``None``.

        :return: Tensor with scores.
        """

    def _validate_candidates_to_score(self, candidates: torch.LongTensor) -> None:
        """Check if candidates param has proper type"""

        if not (isinstance(candidates, torch.Tensor) and candidates.dtype is torch.long):
            msg = (
                "Expected candidates to be of type ``torch.Tensor`` with dtype ``torch.long``, "
                f"got {type(candidates)} with dtype {candidates.dtype}."
            )
            raise ValueError(msg)

    def _valilade_predict_input(self, batch: Any, candidates_to_score: Optional[torch.LongTensor] = None) -> None:
        if self._num_candidates_to_score is None and candidates_to_score is not None:
            msg = (
                "If ``num_candidates_to_score`` is None, "
                "it is impossible to infer the model with passed ``candidates_to_score``."
            )
            raise ValueError(msg)

        if self._batch_size != -1 and batch.padding_mask.shape[0] != self._batch_size:
            msg = (
                f"The batch is smaller then defined batch_size={self._batch_size}. "
                "It is impossible to infer the model with dynamic batch size in ``mode`` = ``batch``. "
                "Use ``mode`` = ``dynamic_batch_size``."
            )
            raise ValueError(msg)

    def _set_inner_params_from_openvino_model(self, compiled_model: ov.CompiledModel) -> None:
        """Set params for ``predict`` method"""

        input_scheme = compiled_model.inputs
        self._batch_size = input_scheme[0].partial_shape[0].max_length
        self._max_seq_len = input_scheme[0].partial_shape[1].max_length
        self._inputs_names = [input.names.pop() for input in compiled_model.inputs]
        if "candidates_to_score" in self._inputs_names:
            self._num_candidates_to_score = input_scheme[-1].partial_shape[0].max_length
        else:
            self._num_candidates_to_score = None
        self._output_name = compiled_model.output().names.pop()

    @staticmethod
    def _validate_num_candidates_to_score(
        num_candidates: Union[int, None],
    ) -> Union[int, None]:
        """Check if num_candidates param is proper"""

        if num_candidates is None:
            return num_candidates
        if isinstance(num_candidates, int) and (num_candidates == -1 or num_candidates >= 1):
            return num_candidates

        msg = (
            "Expected num_candidates_to_score to be of type ``int``, equal to ``-1``, ``natural number`` or ``None``. "
            f"Got {num_candidates}."
        )
        raise ValueError(msg)

    @staticmethod
    def _get_input_params(
        mode: OptimizedModeType,
        batch_size: Optional[int],
        num_candidates_to_score: Optional[int],
    ) -> None:
        """Get params for model compilation according to compilation mode"""

        if mode == "one_query":
            batch_size = 1

        if mode == "batch":
            assert batch_size, f"{mode} mode requires `batch_size`"
            batch_size = batch_size

        if mode == "dynamic_batch_size":
            batch_size = -1

        num_candidates_to_score = num_candidates_to_score if num_candidates_to_score else None
        return batch_size, num_candidates_to_score

    @staticmethod
    def _run_model_compilation(
        lightning_model: lightning.LightningModule,
        model_input_sample: tuple[Union[torch.Tensor, dict[str, torch.Tensor]]],
        model_input_names: list[str],
        model_dynamic_axes_in_input: dict[str, dict],
        batch_size: int,
        num_candidates_to_score: Union[int, None],
        num_threads: Optional[int] = None,
        onnx_path: Optional[str] = None,
    ) -> ov.CompiledModel:
        """
        Model conversion into ONNX format and compilation with defined engine.

        :param lightning_model: Lightning model to be compiled.
        :param model_input_sample: An example of model input with proper data type.
        :param model_input_names: Input tensor names.
        :param model_dynamic_axes_in_input: Dynamic axes in input.
        :param batch_size: Defines the size of the axis with index 0 in the input of the compiled model.
        :param num_candidates_to_score: Defines the size of the candidates in the input of the compiled model.
        :param num_threads: Number of CPU threads to use.
            Must be a natural number or ``None``.
            If ``None``, then compiler will set this parameter automatically.
            Default: ``None``.
        :param onnx_path: Save ONNX model to path, if defined.
            Default: ``None``.

        :return: Compiled model.
        """
        max_seq_len = lightning_model._model.max_len

        if onnx_path is None:
            is_saveble = False
            onnx_file = tempfile.NamedTemporaryFile(suffix=".onnx")
            onnx_path = onnx_file.name
        else:
            is_saveble = True

        lightning_model.to_onnx(
            onnx_path,
            input_sample=model_input_sample,
            export_params=True,
            opset_version=torch.onnx._constants.ONNX_DEFAULT_OPSET,
            do_constant_folding=True,
            input_names=model_input_names,
            output_names=["scores"],
            dynamic_axes=model_dynamic_axes_in_input,
        )
        del lightning_model

        compiled_model = _compile_openvino(onnx_path, batch_size, max_seq_len, num_candidates_to_score, num_threads)

        if not is_saveble:
            onnx_file.close()

        return compiled_model

    @classmethod
    @abstractmethod
    def compile(
        cls,
        model: Union[lightning.LightningModule, str, pathlib.Path],
        mode: OptimizedModeType = "one_query",
        batch_size: Optional[int] = None,
        num_candidates_to_score: Optional[int] = None,
        num_threads: Optional[int] = None,
        onnx_path: Optional[str] = None,
    ) -> "BaseCompiledModel":
        """
        Model compilation.

        :param model: Path to lightning model saved in .ckpt format or the model object itself.
        :param mode: Inference mode, defines shape of inputs.
            Could be one of [``one_query``, ``batch``, ``dynamic_batch_size``].\n
            ``one_query`` - sets input shape to [1, max_seq_len]\n
            ``batch`` - sets input shape to [batch_size, max_seq_len]\n
            ``dynamic_batch_size`` - sets batch_size to dynamic range [?, max_seq_len]\n
            Default: ``one_query``.
        :param batch_size: Batch size, required for ``batch`` mode.
            Default: ``None``.
        :param num_candidates_to_score: Number of item ids to calculate scores.
            Could be one of [``None``, ``-1``, ``N``].\n
            ``-1`` - sets candidates_to_score shape to dynamic range [1, ?]\n
            ``N`` - sets candidates_to_score shape to [1, N]\n
            ``None`` - disable candidates_to_score usage\n
            Default: ``None``.
        :param num_threads: Number of CPU threads to use.
            Must be a natural number or ``None``.
            If ``None``, then compiler will set this parameter automatically.
            Default: ``None``.
        :param onnx_path: Save ONNX model to path, if defined.
            Default: ``None``.
        """



from typing import Literal, Optional

import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_splitter import Splitter, SplitterReturnType

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window

StrategyName = Literal["query"]


class KFolds(Splitter):
    """
    Splits interactions inside each query into folds at random.
    """

    _init_arg_names = [
        "n_folds",
        "strategy",
        "drop_cold_users",
        "drop_cold_items",
        "seed",
        "query_column",
        "item_column",
        "timestamp_column",
        "session_id_column",
        "session_id_processing_strategy",
    ]

    def __init__(
        self,
        n_folds: Optional[int] = 5,
        strategy: Optional[StrategyName] = "query",
        drop_cold_items: bool = False,
        drop_cold_users: bool = False,
        seed: Optional[int] = None,
        query_column: str = "query_id",
        item_column: Optional[str] = "item_id",
        timestamp_column: Optional[str] = "timestamp",
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
    ):
        """
        :param n_folds: number of folds.
        :param strategy: splitting strategy. Only query variant is available atm.
        :param drop_cold_items: flag to drop cold items from test
        :param drop_cold_users: flag to drop cold users from test
        :param seed: random seed
        :param query_column: query id column name
        :param item_column: item id column name
        :param timestamp_column: timestamp column name
        :param session_id_column: name of session id column, which values can not be split.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        super().__init__(
            drop_cold_items=drop_cold_items,
            drop_cold_users=drop_cold_users,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
            session_id_column=session_id_column,
            session_id_processing_strategy=session_id_processing_strategy,
        )
        self.n_folds = n_folds
        if strategy not in {"query"}:
            msg = f"Wrong splitter parameter: {strategy}"
            raise ValueError(msg)
        self.strategy = strategy
        self.seed = seed

    def split(self, interactions: DataFrameLike) -> SplitterReturnType:
        """
        Splits input DataFrame into train and test

        :param interactions: input DataFrame ``[timestamp, user_id, item_id, relevance]``
        :returns: List of splitted DataFrames
        """
        return self._core_split(interactions)

    def _query_split_spark(self, interactions: SparkDataFrame) -> tuple[SparkDataFrame, SparkDataFrame]:
        dataframe = interactions.withColumn("_rand", sf.rand(self.seed))
        dataframe = dataframe.withColumn(
            "fold",
            sf.row_number().over(Window.partitionBy(self.query_column).orderBy("_rand")) % self.n_folds,
        ).drop("_rand")
        for i in range(self.n_folds):
            dataframe = dataframe.withColumn("is_test", sf.when(sf.col("fold") == i, True).otherwise(False))
            if self.session_id_column:
                dataframe = self._recalculate_with_session_id_column(dataframe)

            train = dataframe.filter(~sf.col("is_test")).drop("is_test", "fold")
            test = dataframe.filter(sf.col("is_test")).drop("is_test", "fold")

            test = self._drop_cold_items_and_users(train, test)
            yield train, test

    def _query_split_pandas(self, interactions: PandasDataFrame) -> tuple[PandasDataFrame, PandasDataFrame]:
        dataframe = interactions.sample(frac=1, random_state=self.seed).sort_values(self.query_column)
        dataframe["fold"] = (dataframe.groupby(self.query_column, sort=False).cumcount() + 1) % self.n_folds
        for i in range(self.n_folds):
            dataframe["is_test"] = dataframe["fold"] == i
            if self.session_id_column:
                dataframe = self._recalculate_with_session_id_column(dataframe)

            train = dataframe[~dataframe["is_test"]].drop(columns=["is_test", "fold"])
            test = dataframe[dataframe["is_test"]].drop(columns=["is_test", "fold"])
            dataframe.drop(columns=["is_test"])

            test = self._drop_cold_items_and_users(train, test)
            yield train, test

    def _query_split_polars(self, interactions: PolarsDataFrame) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        dataframe = interactions.sample(fraction=1, shuffle=True, seed=self.seed).sort(self.query_column)
        dataframe = dataframe.with_columns(
            (pl.cum_count(self.query_column).over(self.query_column) % self.n_folds).alias("fold")
        )
        for i in range(self.n_folds):
            dataframe = dataframe.with_columns(
                pl.when(pl.col("fold") == i).then(True).otherwise(False).alias("is_test")
            )
            if self.session_id_column:
                dataframe = self._recalculate_with_session_id_column(dataframe)

            train = dataframe.filter(~pl.col("is_test")).drop("is_test", "fold")
            test = dataframe.filter(pl.col("is_test")).drop("is_test", "fold")

            test = self._drop_cold_items_and_users(train, test)
            yield train, test

    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        if self.strategy == "query":
            if isinstance(interactions, SparkDataFrame):
                return self._query_split_spark(interactions)
            if isinstance(interactions, PandasDataFrame):
                return self._query_split_pandas(interactions)
            if isinstance(interactions, PolarsDataFrame):
                return self._query_split_polars(interactions)

            msg = f"{self} is not implemented for {type(interactions)}"
            raise NotImplementedError(msg)



from collections import OrderedDict
from collections.abc import ItemsView, Iterable, Iterator, KeysView, Mapping, Sequence, ValuesView
from typing import (
    Optional,
    Union,
)

import torch

from replay.data import FeatureHint, FeatureSource, FeatureType

# Alias
TensorMap = Mapping[str, torch.Tensor]
MutableTensorMap = dict[str, torch.Tensor]


class TensorFeatureSource:
    """
    Describes source of a feature
    """

    def __init__(
        self,
        source: FeatureSource,
        column: str,
        index: Optional[int] = None,
    ) -> None:
        """
        :param source: feature source
        :param column: column name
        :param index: index of column in dataframe to get tensor
            directly, without mappings
        """
        self._column = column
        self._index = index
        self._source = source

    @property
    def source(self) -> FeatureSource:
        """
        :returns: feature source
        """
        return self._source

    @property
    def column(self) -> str:
        """
        :returns: column name
        """
        return self._column

    @property
    def index(self) -> Optional[int]:
        """
        :returns: provided index
        """
        return self._index


class TensorFeatureInfo:
    """
    Information about a tensor feature.
    """

    DEFAULT_EMBEDDING_DIM = 64

    def __init__(
        self,
        name: str,
        feature_type: FeatureType,
        is_seq: bool = False,
        feature_hint: Optional[FeatureHint] = None,
        feature_sources: Optional[list[TensorFeatureSource]] = None,
        cardinality: Optional[int] = None,
        padding_value: int = 0,
        embedding_dim: Optional[int] = None,
        tensor_dim: Optional[int] = None,
    ) -> None:
        """
        :param name: name of feature.
        :param feature_type: type of feature.
        :param is_seq: flag that feature is sequential.
            default: ``False``.
        :param feature_hint: hint to models about feature
            (is timestamp, is rating, is query_id, is item_id),
            default: ``None``.
        :param feature_sources: columns names and DataFrames feature came from,
            default: ``None``.
        :param cardinality: cardinality of categorical feature, required for ids columns,
            optional for others,
            default: ``None``.
        :param padding_value: value to pad sequences to desired length
        :param embedding_dim: embedding dimensions of categorical feature,
            default: ``None``.
        :param tensor_dim: tensor dimensions of numerical feature,
            default: ``None``.
        """
        self._name = name
        self._feature_hint = feature_hint
        self._feature_sources = feature_sources
        self._is_seq = is_seq
        self._padding_value = padding_value

        if not isinstance(feature_type, FeatureType):
            msg = "Unknown feature type"
            raise ValueError(msg)
        self._feature_type = feature_type

        if feature_type in [FeatureType.NUMERICAL, FeatureType.NUMERICAL_LIST] and (cardinality or embedding_dim):
            msg = "Cardinality and embedding dimensions are needed only with categorical feature type."
            raise ValueError(msg)
        self._cardinality = cardinality

        if feature_type in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST] and tensor_dim:
            msg = "Tensor dimensions is needed only with numerical feature type."
            raise ValueError(msg)

        if feature_type in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST]:
            self._embedding_dim = embedding_dim or self.DEFAULT_EMBEDDING_DIM
        else:
            self._tensor_dim = tensor_dim

    @property
    def name(self) -> str:
        """
        :returns: The feature name.
        """
        return self._name

    @property
    def feature_type(self) -> FeatureType:
        """
        :returns: The type of feature.
        """
        return self._feature_type

    @property
    def feature_hint(self) -> Optional[FeatureHint]:
        """
        :returns: The feature hint.
        """
        return self._feature_hint

    def _set_feature_hint(self, hint: FeatureHint) -> None:
        self._feature_hint = hint

    @property
    def feature_sources(self) -> Optional[list[TensorFeatureSource]]:
        """
        :returns: List of sources feature came from.
        """
        return self._feature_sources

    def _set_feature_sources(self, sources: list[TensorFeatureSource]) -> None:
        self._feature_sources = sources

    @property
    def feature_source(self) -> Optional[TensorFeatureSource]:
        """
        :returns: Dataframe info of feature.
        """
        source = self.feature_sources
        if not source:
            return None

        if len(source) > 1:
            msg = "Only one element feature sources can be converted to single feature source."
            raise ValueError(msg)
        assert isinstance(self.feature_sources, list)
        return self.feature_sources[0]

    @property
    def is_seq(self) -> bool:
        """
        :returns: Flag that feature is sequential.\n
        Sequential means that the value of the feature will be determined for each element of the user's sequence.
        """
        return self._is_seq

    @property
    def is_cat(self) -> bool:
        """
        :returns: Flag that feature is categorical.
        """
        return self.feature_type in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST]

    @property
    def is_num(self) -> bool:
        """
        :returns: Flag that feature is numerical.
        """
        return self.feature_type in [FeatureType.NUMERICAL, FeatureType.NUMERICAL_LIST]

    @property
    def is_list(self) -> bool:
        """
        :returns: Flag that feature is numerical list or categorical list.
        """
        return self.feature_type in [FeatureType.CATEGORICAL_LIST, FeatureType.NUMERICAL_LIST]

    @property
    def padding_value(self) -> int:
        """
        :returns: value to pad sequences to desired length.
        """
        return self._padding_value

    @property
    def cardinality(self) -> Optional[int]:
        """
        :returns: Cardinality of the feature.
        """
        if not self.is_cat:
            msg = f"Can not get cardinality because feature type of {self.name} column is not categorical."
            raise RuntimeError(msg)
        return self._cardinality

    def _set_cardinality(self, cardinality: int) -> None:
        self._cardinality = cardinality

    @property
    def tensor_dim(self) -> Optional[int]:
        """
        :returns: Dimensions of the numerical feature.
        """
        if not self.is_num:
            msg = f"Can not get tensor dimensions because feature type of {self.name} feature is not numerical."
            raise RuntimeError(msg)
        return self._tensor_dim

    def _set_tensor_dim(self, tensor_dim: int) -> None:
        self._tensor_dim = tensor_dim

    @property
    def embedding_dim(self) -> Optional[int]:
        """
        :returns: Embedding dimensions of the feature.
        """
        if not self.is_cat:
            msg = f"Can not get embedding dimensions because feature type of {self.name} feature is not categorical."
            raise RuntimeError(msg)
        return self._embedding_dim

    def _set_embedding_dim(self, embedding_dim: int) -> None:
        self._embedding_dim = embedding_dim


class TensorSchema(Mapping[str, TensorFeatureInfo]):
    """
    Key-value like collection that stores tensor features
    """

    def __init__(self, features_list: Union[Sequence[TensorFeatureInfo], TensorFeatureInfo]) -> None:
        """
        :param features_list: list of tensor feature infos.
        """
        if isinstance(features_list, OrderedDict):
            features_list = list(features_list.values())
        features_list = [features_list] if not isinstance(features_list, Sequence) else features_list
        self._tensor_schema = {feature.name: feature for feature in features_list}

    def subset(self, features_to_keep: Iterable[str]) -> "TensorSchema":
        """Creates a subset of given features.

        :param features_to_keep: A sequence of feature names
                in original schema to keep in subset.

        :returns: New tensor schema of given features.
        """
        features: set[TensorFeatureInfo] = set()
        for feature_name in features_to_keep:
            features.add(self._tensor_schema[feature_name])
        return TensorSchema(list(features))

    def item(self) -> TensorFeatureInfo:
        """
        :returns: Extract single feature from a schema.
        """
        if len(self._tensor_schema) != 1:
            msg = "Only one element tensor schema can be converted to single feature"
            raise ValueError(msg)
        return next(iter(self._tensor_schema.values()))

    def items(self) -> ItemsView[str, TensorFeatureInfo]:
        return self._tensor_schema.items()

    def keys(self) -> KeysView[str]:
        return self._tensor_schema.keys()

    def values(self) -> ValuesView[TensorFeatureInfo]:
        return self._tensor_schema.values()

    def get(
        self,
        key: str,
        default: Optional[TensorFeatureInfo] = None,
    ) -> Optional[TensorFeatureInfo]:
        return self._tensor_schema.get(key, default)

    def __iter__(self) -> Iterator[str]:
        return iter(self._tensor_schema)

    def __contains__(self, feature_name: object) -> bool:
        return feature_name in self._tensor_schema

    def __len__(self) -> int:
        return len(self._tensor_schema)

    def __getitem__(self, feature_name: str) -> TensorFeatureInfo:
        return self._tensor_schema[feature_name]

    def __eq__(self, other: object) -> bool:
        return self._tensor_schema == other

    def __ne__(self, other: object) -> bool:
        return self._tensor_schema != other

    def __add__(self, other: "TensorSchema") -> "TensorSchema":
        return TensorSchema(list(self._tensor_schema.values()) + list(other._tensor_schema.values()))

    @property
    def all_features(self) -> Sequence[TensorFeatureInfo]:
        """
        :returns: Sequence of all features.
        """
        return list(self._tensor_schema.values())

    @property
    def categorical_features(self) -> "TensorSchema":
        """
        :returns: Sequence of categorical features in a schema.
        """
        return self.filter(feature_type=FeatureType.CATEGORICAL) + self.filter(
            feature_type=FeatureType.CATEGORICAL_LIST
        )

    @property
    def numerical_features(self) -> "TensorSchema":
        """
        :returns: Sequence of numerical features in a schema.
        """
        return self.filter(feature_type=FeatureType.NUMERICAL) + self.filter(feature_type=FeatureType.NUMERICAL_LIST)

    @property
    def query_id_features(self) -> "TensorSchema":
        """
        :returns: Sequence of query id features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.QUERY_ID)

    @property
    def item_id_features(self) -> "TensorSchema":
        """
        :returns: Sequence of item id features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.ITEM_ID)

    @property
    def timestamp_features(self) -> "TensorSchema":
        """
        :returns: Sequence of timestamp features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.TIMESTAMP)

    @property
    def rating_features(self) -> "TensorSchema":
        """
        :returns: Sequence of rating features in a schema.
        """
        return self.filter(feature_hint=FeatureHint.RATING)

    @property
    def sequential_features(self) -> "TensorSchema":
        """
        :returns: Sequence of sequential features in a schema.
        """
        return self.filter(is_seq=True)

    @property
    def names(self) -> Sequence[str]:
        """
        :returns: List of all feature's names.
        """
        return list(self._tensor_schema)

    @property
    def query_id_feature_name(self) -> Optional[str]:
        """
        :returns: Query id feature name.
        """
        query_id_features = self.query_id_features
        if not query_id_features:
            return None
        return query_id_features.item().name

    @property
    def item_id_feature_name(self) -> Optional[str]:
        """
        :returns: Item id feature name.
        """
        item_id_features = self.item_id_features
        if not item_id_features:
            return None
        return item_id_features.item().name

    @property
    def timestamp_feature_name(self) -> Optional[str]:
        """
        :returns: Timestamp feature name.
        """
        timestamp_features = self.timestamp_features
        if not timestamp_features:
            return None
        return timestamp_features.item().name

    @property
    def rating_feature_name(self) -> Optional[str]:
        """
        :returns: Rating feature name.
        """
        rating_features = self.rating_features
        if not rating_features:
            return None
        return rating_features.item().name

    def _get_object_args(self) -> dict:
        """
        Returns list of features represented as dictionaries.
        """
        features = [
            {
                "name": feature.name,
                "feature_type": feature.feature_type.name,
                "is_seq": feature.is_seq,
                "feature_hint": feature.feature_hint.name if feature.feature_hint else None,
                "feature_sources": (
                    [{"source": x.source.name, "column": x.column, "index": x.index} for x in feature.feature_sources]
                    if feature.feature_sources
                    else None
                ),
                "cardinality": feature.cardinality if feature.is_cat else None,
                "embedding_dim": feature.embedding_dim if feature.is_cat else None,
                "tensor_dim": feature.tensor_dim if feature.is_num else None,
            }
            for feature in self.all_features
        ]
        return features

    @classmethod
    def _create_object_by_args(cls, args: dict) -> "TensorSchema":
        features_list = []
        for feature_data in args:
            feature_data["feature_sources"] = (
                [
                    TensorFeatureSource(source=FeatureSource[x["source"]], column=x["column"], index=x["index"])
                    for x in feature_data["feature_sources"]
                ]
                if feature_data["feature_sources"]
                else None
            )
            f_type = feature_data["feature_type"]
            f_hint = feature_data["feature_hint"]
            feature_data["feature_type"] = FeatureType[f_type] if f_type else None
            feature_data["feature_hint"] = FeatureHint[f_hint] if f_hint else None
            features_list.append(TensorFeatureInfo(**feature_data))
        return TensorSchema(features_list)

    def filter(
        self,
        name: Optional[str] = None,
        feature_hint: Optional[FeatureHint] = None,
        is_seq: Optional[bool] = None,
        feature_type: Optional[FeatureType] = None,
    ) -> "TensorSchema":
        """Filter list by ``name``, ``feature_type``, ``is_seq`` and ``feature_hint``.

        :param name: Feature name to filter by.
            default: ``None``.
        :param feature_hint: Feature hint to filter by.
            default: ``None``.
        :param feature_source: Feature source to filter by.
            default: ``None``.
        :param feature_type: Feature type to filter by.
            default: ``None``.

        :returns: New filtered feature schema.
        """
        filtered_features = self.all_features
        filter_functions = [self._name_filter, self._seq_filter, self._type_filter, self._hint_filter]
        filter_parameters = [name, is_seq, feature_type, feature_hint]
        for filtration_func, filtration_param in zip(filter_functions, filter_parameters):
            filtered_features = list(
                filter(
                    lambda x: filtration_func(x, filtration_param),
                    filtered_features,
                )
            )
        return TensorSchema(filtered_features)

    @staticmethod
    def _name_filter(value: TensorFeatureInfo, name: str) -> bool:
        return value.name == name if name else True

    @staticmethod
    def _seq_filter(value: TensorFeatureInfo, is_seq: bool) -> bool:
        return value.is_seq == is_seq if is_seq is not None else True

    @staticmethod
    def _type_filter(value: TensorFeatureInfo, feature_type: FeatureType) -> bool:
        return value.feature_type == feature_type if feature_type else True

    @staticmethod
    def _hint_filter(value: TensorFeatureInfo, feature_hint: FeatureHint) -> bool:
        return value.feature_hint == feature_hint if feature_hint else True



from abc import ABC, abstractmethod

from replay.models.extensions.ann.entities.base_hnsw_param import BaseHnswParam
from replay.models.extensions.ann.index_stores.base_index_store import IndexStore
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class IndexInferer(ABC):
    """Abstract base class that describes a common interface for index inferers
    and provides common methods for them."""

    # All implementations use the same udf return type.
    udf_return_type = "item_idx array<int>, distance array<double>"

    def __init__(self, index_params: BaseHnswParam, index_store: IndexStore):
        self.index_params = index_params
        self.index_store = index_store

    @abstractmethod
    def infer(self, vectors: SparkDataFrame, features_col: str, k: int) -> SparkDataFrame:
        """Infers index"""

    @staticmethod
    def _unpack_infer_struct(inference_result: SparkDataFrame) -> SparkDataFrame:
        """Transforms input dataframe.
        Unpacks and explodes arrays from `neighbours` struct.

        >>>
        >> inference_result.printSchema()
        root
         |-- user_idx: integer (nullable = true)
         |-- neighbours: struct (nullable = true)
         |    |-- item_idx: array (nullable = true)
         |    |    |-- element: integer (containsNull = true)
         |    |-- distance: array (nullable = true)
         |    |    |-- element: double (containsNull = true)
        >> ANNMixin._unpack_infer_struct(inference_result).printSchema()
        root
         |-- user_idx: integer (nullable = true)
         |-- item_idx: integer (nullable = true)
         |-- relevance: double (nullable = true)

        Args:
            inference_result: output of infer_index UDF
        """
        res = inference_result.select(
            "user_idx",
            sf.explode(sf.arrays_zip("neighbours.item_idx", "neighbours.distance")).alias("zip_exp"),
        )

        # Fix arrays_zip random behavior.
        # It can return zip_exp.0 or zip_exp.item_idx in different machines.
        fields = res.schema["zip_exp"].jsonValue()["type"]["fields"]
        item_idx_field_name: str = fields[0]["name"]
        distance_field_name: str = fields[1]["name"]

        res = res.select(
            "user_idx",
            sf.col(f"zip_exp.{item_idx_field_name}").alias("item_idx"),
            (sf.lit(-1.0) * sf.col(f"zip_exp.{distance_field_name}")).alias("relevance"),
        )
        return res



import numpy as np
import pandas as pd

from replay.models.extensions.ann.utils import create_hnswlib_index_instance
from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, SparkDataFrame
from replay.utils.session_handler import State

from .base_inferer import IndexInferer

if PYSPARK_AVAILABLE:
    from pyspark.sql.pandas.functions import pandas_udf


class HnswlibFilterIndexInferer(IndexInferer):
    """Hnswlib index inferer with filter seen items. Infers hnswlib index."""

    def infer(self, vectors: SparkDataFrame, features_col: str, k: int) -> SparkDataFrame:
        _index_store = self.index_store
        index_params = self.index_params

        index_store_broadcast = State().session.sparkContext.broadcast(_index_store)

        @pandas_udf(self.udf_return_type)
        def infer_index_udf(
            vectors: pd.Series,
            num_items: pd.Series,
            seen_item_ids: pd.Series,
        ) -> PandasDataFrame:  # pragma: no cover
            index_store = index_store_broadcast.value
            index = index_store.load_index(
                init_index=lambda: create_hnswlib_index_instance(index_params),
                load_index=lambda index, path: index.load_index(path),
                configure_index=lambda index: index.set_ef(index_params.ef_s) if index_params.ef_s else None,
            )

            # max number of items to retrieve per batch
            max_items_to_retrieve = num_items.max()

            labels, distances = index.knn_query(
                np.stack(vectors.values),
                k=k + max_items_to_retrieve,
                num_threads=1,
            )

            filtered_labels = []
            filtered_distances = []
            for i, item_ids in enumerate(labels):
                non_seen_item_indexes = ~np.isin(item_ids, seen_item_ids[i], assume_unique=True)
                filtered_labels.append((item_ids[non_seen_item_indexes])[:k])
                filtered_distances.append((distances[i][non_seen_item_indexes])[:k])

            pd_res = pd.DataFrame(
                {
                    "item_idx": filtered_labels,
                    "distance": filtered_distances,
                }
            )

            return pd_res

        cols = ["num_items", "seen_item_idxs"]

        res = vectors.select(
            "user_idx",
            infer_index_udf(features_col, *cols).alias("neighbours"),
        )
        res = self._unpack_infer_struct(res)

        return res



from collections.abc import Iterable
from os.path import join
from typing import Optional, Union

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import IsSavable
from .common import RecommenderCommons

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf

    from replay.utils.spark_utils import (
        filter_cold,
        get_top_k,
        get_unique_entities,
        load_pickled_from_parquet,
        return_recs,
        save_picklable_to_parquet,
    )


class CatPopRec(IsSavable, RecommenderCommons):
    """
    CatPopRec generate recommendation for item categories.
    It recommends the most popular items in the category or all sub-categories in the category.
    It is assumed that each item belongs to one or more categories.
    Categories could have a flat structure or be hierarchical (tree structure).
    If the categories are hierarchical, items should belong to the lowest level categories only.
    In this case, the model could generate recommendations not only for the lowest level categories,
    but for all categories, combining the lowest level categories statistics.
    """

    cat_item_popularity: SparkDataFrame
    leaf_cat_mapping: SparkDataFrame
    can_predict_cold_items: bool = False
    fit_items: SparkDataFrame

    def _generate_mapping(self, cat_tree: SparkDataFrame, max_iter: int = 20) -> SparkDataFrame:
        """
        Create SparkDataFrame with mapping [`category`, `leaf_cat`]
        where `leaf_cat` is the lowest level categories of category tree,
        which contain items, not sub-categories.
        :param cat_tree: spark dataframe with columns [`category`, `parent_cat`].
            Contains category mapping `category - parent category`.
            Each category has only one parent.
            If the parent is absent, `parent_cat` value should be None.
        :param max_iter: maximal number of iteration of descend through the category tree
        :return: SparkDataFrame with mapping [`category`, `leaf_cat`]
        """
        current_res = cat_tree.select(sf.col("category"), sf.col("category").alias("leaf_cat"))

        i = 0
        res_size_growth = current_res.count()
        while i < max_iter and res_size_growth > 0:
            new_res = (
                current_res.join(
                    cat_tree.withColumnRenamed("category", "new_leaf_cat"),
                    on=sf.col("leaf_cat") == sf.col("parent_cat"),
                    how="left",
                )
                # if `leaf_cat` is already found, the join result will be None,
                # use coalesce to get value
                .select(
                    "category",
                    sf.coalesce("new_leaf_cat", "leaf_cat").alias("leaf_cat"),
                )
            ).cache()
            res_size_growth = new_res.count() - current_res.count()
            current_res.unpersist()
            current_res = new_res
            i += 1

        # RunTime error instead of logger.warning?
        if i == max_iter:
            self.logger.warning(
                "Category tree was not fully processed in %s iterations. "
                "Increase the `max_iter` value or check the category tree structure."
                "It must not have loops and each category should have only one parent.",
                max_iter,
            )

        return current_res

    def set_cat_tree(self, cat_tree: SparkDataFrame):
        """
        Set/update category tree `cat_tree` used to generate recommendations.
        :param cat_tree: park dataframe with columns [`category`, `parent_cat`].
            Contains category mapping `category - parent category`.
            Each category has only one parent.
        """
        self._clear_cache()
        self.leaf_cat_mapping = self._generate_mapping(cat_tree)

    def __init__(
        self,
        cat_tree: Optional[SparkDataFrame] = None,
        max_iter: Optional[int] = 20,
    ):
        """
        :param cat_tree: spark dataframe with columns [`category`, `parent_cat`].
            Contains category mapping `category - parent category`.
            Each category has only one parent.
        :param max_iter: maximal number of iteration of descend through the category tree
        """
        self.max_iter = max_iter
        if cat_tree is not None:
            self.leaf_cat_mapping = self._generate_mapping(cat_tree, max_iter=max_iter)

    @property
    def _init_args(self):
        return {"max_iter": self.max_iter}

    @property
    def _dataframes(self):
        return {
            "cat_item_popularity": self.cat_item_popularity,
            "leaf_cat_mapping": self.leaf_cat_mapping,
        }

    def fit(self, dataset: Dataset) -> None:
        """
        Fit a recommendation model

        :param dataset: historical interactions with query/item features
            ``[user_id, item_id, category, timestamp, rating]``
            where `category` is an item's category.
            The `item_idx`, `category` are mandatory columns.
            If `rating` column is present it is treated as number
            of interactions with the item and the `rating` values are summed.
        """
        self.logger.debug("Starting fit %s", type(self).__name__)
        self.query_column = dataset.feature_schema.query_id_column
        self.item_column = dataset.feature_schema.item_id_column
        self.rating_column = dataset.feature_schema.interactions_rating_column

        self.fit_items = sf.broadcast(dataset.interactions.select(self.item_column).distinct())
        self._fit(dataset=dataset)

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        if self.rating_column is not None:
            self.cat_item_popularity = dataset.interactions.groupBy("category", self.item_column).agg(
                sf.sum(self.rating_column).alias(self.rating_column)
            )
        else:
            self.rating_column = "relevance"
            self.cat_item_popularity = dataset.interactions.groupBy("category", self.item_column).agg(
                sf.count(self.item_column).alias(self.rating_column)
            )

        self.cat_item_popularity.cache()
        self.cat_item_popularity.count()

    def _clear_cache(self):
        if hasattr(self, "cat_item_popularity"):
            self.cat_item_popularity.unpersist()
        if hasattr(self, "leaf_cat_mapping"):
            self.leaf_cat_mapping.unpersist()

    def predict(
        self,
        categories: Union[SparkDataFrame, Iterable],
        k: int,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get top-k recommendations for each category in `categories`.

        :param categories:  dataframe containing ``[category]`` to predict for or ``array-like``;
        :param k: number of recommendations for each category
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, consider all items.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe will be returned
        :return: cached recommendation dataframe with columns ``[category, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._predict_wrap(
            categories=categories,
            k=k,
            items=items,
            recs_file_path=recs_file_path,
        )

    def _predict_wrap(
        self,
        categories: Union[SparkDataFrame, Iterable],
        k: int,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Predict wrapper to allow for fewer parameters in models

        :param categories:  dataframe containing ``[category]`` to predict for or ``array-like``;
        :param k: number of recommendations for each user
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items presented in `fit` data.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[category, item_idx, rating]``
            or None if `file_path` is provided
        """
        self.logger.debug("Starting predict %s", type(self).__name__)

        categories = get_unique_entities(categories, "category")
        item_data = items or self.fit_items
        items = get_unique_entities(item_data, self.item_column)

        num_new, items = filter_cold(items, self.fit_items, col_name=self.item_column)
        if num_new > 0:
            self.logger.info(
                "%s model can't predict cold items, they will be ignored",
                self,
            )

        num_items = items.count()
        if num_items < k:
            message = f"k = {k} > number of items = {num_items}"
            self.logger.debug(message)

        recs = self._predict(
            categories=categories,
            items=items,
        )

        recs = get_top_k(
            recs,
            k=k,
            partition_by_col=sf.col("category"),
            order_by_col=[
                sf.col(self.rating_column).desc(),
                sf.col(self.item_column).desc(),
            ],
        ).select("category", self.item_column, self.rating_column)

        return return_recs(recs, recs_file_path)

    def _predict(
        self,
        categories: Union[SparkDataFrame, Iterable],
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
    ) -> SparkDataFrame:
        res = categories.join(self.leaf_cat_mapping, on="category")
        # filter required categories and items of `self.cat_item_popularity`
        unique_leaf_cats = res.select("leaf_cat").distinct()
        unique_leaf_cat_items = (
            self.cat_item_popularity.withColumnRenamed("category", "leaf_cat")
            .join(sf.broadcast(unique_leaf_cats), on="leaf_cat")
            .join(sf.broadcast(items), on=self.item_column)
        )

        # find number of interactions in all leaf categories after filtering
        num_interactions_in_cat = (
            res.join(
                unique_leaf_cat_items.groupBy("leaf_cat").agg(sf.sum(self.rating_column).alias("sum_rating")),
                on="leaf_cat",
            )
            .groupBy("category")
            .agg(sf.sum("sum_rating").alias("sum_rating"))
        )

        # aggregate results for each category: sum up num interactions in leaf categories
        # and calculate popularity as a number of interactions with an item in all leaf categories
        # divided by the number of interactions in all leaf categories of a category
        return (
            unique_leaf_cat_items.join(res, on="leaf_cat")
            .groupBy("category", self.item_column)
            .agg(sf.sum(self.rating_column).alias(self.rating_column))
            .join(num_interactions_in_cat, on="category")
            .withColumn(self.rating_column, sf.col(self.rating_column) / sf.col("sum_rating"))
        )

    def _save_model(self, path: str):
        save_picklable_to_parquet(
            {
                "query_column": self.query_column,
                "item_column": self.item_column,
                "rating_column": self.rating_column,
            },
            join(path, "params.dump"),
        )

    def _load_model(self, path: str):
        loaded_params = load_pickled_from_parquet(join(path, "params.dump"))
        for param, value in loaded_params.items():
            setattr(self, param, value)



from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import Recommender

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class QueryPopRec(Recommender):
    """
    Recommends old objects from each query's personal top.
    Input is the number of interactions between queries and items.

    Popularity for item :math:`i` and query :math:`u` is defined as the
    fraction of actions with item :math:`i` among all interactions of query :math:`u`:

    .. math::
        Popularity(i_u) = \\dfrac{N_iu}{N_u}

    :math:`N_iu` - number of interactions of query :math:`u` with item :math:`i`.
    :math:`N_u` - total number of interactions of query :math:`u`.

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame({"user_id": [1, 1, 3], "item_id": [1, 2, 3], "rating": [2, 1, 1]})
    >>> data_frame
        user_id   item_id     rating
    0         1         1          2
    1         1         2          1
    2         3         3          1

    >>> interactions = convert2spark(data_frame)
    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> dataset = Dataset(feature_schema, interactions)
    >>> model = QueryPopRec()
    >>> res = model.fit_predict(dataset, 1, filter_seen_items=False)
    >>> model.query_item_popularity.count()
    3
    >>> res.toPandas().sort_values("user_id", ignore_index=True)
        user_id   item_id     rating
    0         1         1   0.666667
    1         3         3   1.000000
    """

    query_item_popularity: SparkDataFrame

    @property
    def _init_args(self):
        return {}

    @property
    def _dataframes(self):
        return {"query_item_popularity": self.query_item_popularity}

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        query_rating_sum = (
            dataset.interactions.groupBy(self.query_column)
            .agg(sf.sum(self.rating_column).alias("query_rel_sum"))
            .withColumnRenamed(self.query_column, "query")
            .select("query", "query_rel_sum")
        )
        self.query_item_popularity = (
            dataset.interactions.groupBy(self.query_column, self.item_column)
            .agg(sf.sum(self.rating_column).alias("query_item_rel_sum"))
            .join(
                query_rating_sum,
                how="inner",
                on=sf.col(self.query_column) == sf.col("query"),
            )
            .select(
                self.query_column,
                self.item_column,
                (sf.col("query_item_rel_sum") / sf.col("query_rel_sum")).alias(self.rating_column),
            )
        )
        self.query_item_popularity.cache().count()

    def _clear_cache(self):
        if hasattr(self, "query_item_popularity"):
            self.query_item_popularity.unpersist()

    def _predict(
        self,
        dataset: Dataset,  # noqa: ARG002
        k: int,  # noqa: ARG002
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        if filter_seen_items:
            self.logger.warning("QueryPopRec can't predict new items, recommendations will not be filtered")

        return self.query_item_popularity.join(queries, on=self.query_column).join(items, on=self.item_column)



from .dataset import (
    SasRecPredictionBatch,
    SasRecPredictionDataset,
    SasRecTrainingBatch,
    SasRecTrainingDataset,
    SasRecValidationBatch,
    SasRecValidationDataset,
)
from .lightning import SasRec
from .model import SasRecModel



from typing import Optional

import polars as pl

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

from .base_splitter import Splitter, SplitterReturnType

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf


class ColdUserRandomSplitter(Splitter):
    """
    Test set consists of all actions of randomly chosen users.
    """

    _init_arg_names = [
        "test_size",
        "drop_cold_items",
        "seed",
        "query_column",
        "item_column",
    ]

    def __init__(
        self,
        test_size: float,
        drop_cold_items: bool = False,
        seed: Optional[int] = None,
        query_column: str = "query_id",
        item_column: Optional[str] = "item_id",
    ):
        """
        :param test_size: The proportion of users to allocate to the test set.
            Must be a float between 0.0 and 1.0.
        :param drop_cold_items: Drop items from test DataFrame
            which are not in train DataFrame, default: False.
        :param seed: Seed for the random number generator to ensure
            reproducibility of the split, default: None.
        :param query_column: Name of query interaction column.
            default: ``query_id``.
        :param item_column: Name of item interaction column.
            default: ``item_id``.
        """
        super().__init__(
            drop_cold_items=drop_cold_items,
            query_column=query_column,
            item_column=item_column,
        )
        self.seed = seed
        if test_size <= 0 or test_size >= 1:
            msg = "test_size must between 0 and 1"
            raise ValueError(msg)
        self.test_size = test_size

    def _core_split_pandas(
        self, interactions: PandasDataFrame, threshold: float
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        users = PandasDataFrame(interactions[self.query_column].unique(), columns=[self.query_column])
        train_users = users.sample(frac=(1 - threshold), random_state=self.seed)
        train_users["is_test"] = False

        interactions = interactions.merge(train_users, on=self.query_column, how="left")
        interactions["is_test"].fillna(True, inplace=True)

        train = interactions[~interactions["is_test"]].drop(columns=["is_test"])
        test = interactions[interactions["is_test"]].drop(columns=["is_test"])
        interactions = interactions.drop(columns=["is_test"])

        return train, test

    def _core_split_spark(
        self, interactions: SparkDataFrame, threshold: float
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        users = interactions.select(self.query_column).distinct()
        train_users, _ = users.randomSplit(
            [1 - threshold, threshold],
            seed=self.seed,
        )
        interactions = interactions.join(
            train_users.withColumn("is_test", sf.lit(False)),
            on=self.query_column,
            how="left",
        ).na.fill({"is_test": True})

        train = interactions.filter(~sf.col("is_test")).drop("is_test")
        test = interactions.filter(sf.col("is_test")).drop("is_test")

        return train, test

    def _core_split_polars(
        self, interactions: PolarsDataFrame, threshold: float
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        train_users = (
            interactions.select(self.query_column)
            .unique(maintain_order=True)
            .sample(fraction=(1 - threshold), seed=self.seed)
            .with_columns(pl.lit(False).alias("is_test"))
        )

        interactions = interactions.join(train_users, on=self.query_column, how="left").fill_null(True)

        train = interactions.filter(~pl.col("is_test")).drop("is_test")
        test = interactions.filter(pl.col("is_test")).drop("is_test")
        return train, test

    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        if isinstance(interactions, SparkDataFrame):
            return self._core_split_spark(interactions, self.test_size)
        if isinstance(interactions, PandasDataFrame):
            return self._core_split_pandas(interactions, self.test_size)
        if isinstance(interactions, PolarsDataFrame):
            return self._core_split_polars(interactions, self.test_size)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)



from .dataset import Dataset
from .schema import FeatureHint, FeatureInfo, FeatureSchema, FeatureSource, FeatureType
from .spark_schema import get_schema

__all__ = [
    "Dataset",
    "FeatureHint",
    "FeatureInfo",
    "FeatureSchema",
    "FeatureSource",
    "FeatureType",
    "get_schema",
]



from typing import Optional

from replay.data import Dataset
from replay.utils import OPTUNA_AVAILABLE, PYSPARK_AVAILABLE, SparkDataFrame

from .base_neighbour_rec import NeighbourRec
from .extensions.ann.index_builders.base_index_builder import IndexBuilder

if OPTUNA_AVAILABLE:
    from replay.models.optimization import ItemKNNObjective

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf
    from pyspark.sql.window import Window


class ItemKNN(NeighbourRec):
    """Item-based ItemKNN with modified cosine similarity measure."""

    def _get_ann_infer_params(self) -> dict:
        return {
            "features_col": None,
        }

    all_items: Optional[SparkDataFrame]
    dot_products: Optional[SparkDataFrame]
    item_norms: Optional[SparkDataFrame]
    bm25_k1 = 1.2
    bm25_b = 0.75

    _valid_weightings = [None, "tf_idf", "bm25"]
    if OPTUNA_AVAILABLE:
        _objective = ItemKNNObjective
        _search_space = {
            "num_neighbours": {"type": "int", "args": [1, 100]},
            "shrink": {"type": "int", "args": [0, 100]},
            "weighting": {"type": "categorical", "args": _valid_weightings},
        }

    def __init__(
        self,
        num_neighbours: int = 10,
        use_rating: bool = False,
        shrink: float = 0.0,
        weighting: Optional[str] = None,
        index_builder: Optional[IndexBuilder] = None,
    ):
        """
        :param num_neighbours: number of neighbours
        :param use_rating: flag to use rating values as is or to treat them as 1
        :param shrink: term added to the denominator when calculating similarity
        :param weighting: item reweighting type, one of [None, 'tf_idf', 'bm25']
        :param index_builder: `IndexBuilder` instance that adds ANN functionality.
            If not set, then ann will not be used.
        """
        self.init_index_builder(index_builder)
        self.shrink = shrink
        self.use_rating = use_rating
        self.num_neighbours = num_neighbours

        if weighting not in self._valid_weightings:
            msg = f"weighting must be one of {self._valid_weightings}"
            raise ValueError(msg)
        self.weighting = weighting

    @property
    def _init_args(self):
        return {
            "shrink": self.shrink,
            "use_rating": self.use_rating,
            "num_neighbours": self.num_neighbours,
            "weighting": self.weighting,
            "index_builder": self.index_builder.init_meta_as_dict() if self.index_builder else None,
        }

    @staticmethod
    def _shrink(dot_products: SparkDataFrame, shrink: float) -> SparkDataFrame:
        return dot_products.withColumn(
            "similarity",
            sf.col("dot_product") / (sf.col("norm1") * sf.col("norm2") + shrink),
        ).select("item_idx_one", "item_idx_two", "similarity")

    def _get_similarity(self, interactions: SparkDataFrame) -> SparkDataFrame:
        """
        Calculate item similarities

        :param interactions: SparkDataFrame with interactions, `[user_id, item_id, rating]`
        :return: similarity matrix `[item_idx_one, item_idx_two, similarity]`
        """
        dot_products = self._get_products(interactions)
        similarity = self._shrink(dot_products, self.shrink)
        return similarity

    def _reweight_interactions(self, interactions: SparkDataFrame):
        """
        Reweight rating according to TD-IDF or BM25 weighting.

        :param interactions: SparkDataFrame with interactions, `[user_id, item_id, rating]`
        :return: interactions `[user_id, item_id, rating]`
        """
        if self.weighting == "bm25":
            interactions = self._get_tf_bm25(interactions)

        idf = self._get_idf(interactions)

        interactions = interactions.join(idf, how="inner", on=self.query_column).withColumn(
            self.rating_column,
            sf.col(self.rating_column) * sf.col("idf"),
        )

        return interactions

    def _get_tf_bm25(self, interactions: SparkDataFrame):
        """
        Adjust rating by BM25 term frequency.

        :param interactions: SparkDataFrame with interactions, `[user_id, item_id, rating]`
        :return: interactions `[user_id, item_id, rating]`
        """
        item_stats = interactions.groupBy(self.item_column).agg(sf.count(self.query_column).alias("n_queries_per_item"))
        avgdl = item_stats.select(sf.mean("n_queries_per_item")).take(1)[0][0]
        interactions = interactions.join(item_stats, how="inner", on=self.item_column)

        interactions = interactions.withColumn(
            self.rating_column,
            sf.col(self.rating_column)
            * (self.bm25_k1 + 1)
            / (
                sf.col(self.rating_column)
                + self.bm25_k1 * (1 - self.bm25_b + self.bm25_b * (sf.col("n_queries_per_item") / avgdl))
            ),
        ).drop("n_queries_per_item")

        return interactions

    def _get_idf(self, interactions: SparkDataFrame):
        """
        Return inverse document score for interactions reweighting.

        :param interactions: SparkDataFrame with interactions, `[user_id, item_id, rating]`
        :return: idf `[idf]`
        :raises: ValueError if self.weighting not in ["tf_idf", "bm25"]
        """
        df = interactions.groupBy(self.query_column).agg(sf.count(self.item_column).alias("DF"))
        n_items = interactions.select(self.item_column).distinct().count()

        if self.weighting == "tf_idf":
            idf = df.withColumn("idf", sf.log1p(sf.lit(n_items) / sf.col("DF"))).drop("DF")
        elif self.weighting == "bm25":
            idf = df.withColumn(
                "idf",
                sf.log1p((sf.lit(n_items) - sf.col("DF") + 0.5) / (sf.col("DF") + 0.5)),
            ).drop("DF")
        else:
            msg = "weighting must be one of ['tf_idf', 'bm25']"
            raise ValueError(msg)

        return idf

    def _get_products(self, interactions: SparkDataFrame) -> SparkDataFrame:
        """
        Calculate item dot products

        :param interactions: SparkDataFrame with interactions, `[user_id, item_id, rating]`
        :return: similarity matrix `[item_idx_one, item_idx_two, norm1, norm2]`
        """
        if self.weighting:
            interactions = self._reweight_interactions(interactions)

        left = interactions.withColumnRenamed(self.item_column, "item_idx_one").withColumnRenamed(
            self.rating_column, "rel_one"
        )
        right = interactions.withColumnRenamed(self.item_column, "item_idx_two").withColumnRenamed(
            self.rating_column, "rel_two"
        )

        dot_products = (
            left.join(right, how="inner", on=self.query_column)
            .filter(sf.col("item_idx_one") != sf.col("item_idx_two"))
            .withColumn(self.rating_column, sf.col("rel_one") * sf.col("rel_two"))
            .groupBy("item_idx_one", "item_idx_two")
            .agg(sf.sum(self.rating_column).alias("dot_product"))
        )

        item_norms = (
            interactions.withColumn(self.rating_column, sf.col(self.rating_column) ** 2)
            .groupBy(self.item_column)
            .agg(sf.sum(self.rating_column).alias("square_norm"))
            .select(sf.col(self.item_column), sf.sqrt("square_norm").alias("norm"))
        )
        norm1 = item_norms.withColumnRenamed(self.item_column, "item_id1").withColumnRenamed("norm", "norm1")
        norm2 = item_norms.withColumnRenamed(self.item_column, "item_id2").withColumnRenamed("norm", "norm2")

        dot_products = dot_products.join(norm1, how="inner", on=sf.col("item_id1") == sf.col("item_idx_one"))
        dot_products = dot_products.join(norm2, how="inner", on=sf.col("item_id2") == sf.col("item_idx_two"))

        return dot_products

    def _get_k_most_similar(self, similarity_matrix: SparkDataFrame) -> SparkDataFrame:
        """
        Leaves only top-k neighbours for each item

        :param similarity_matrix: dataframe `[item_idx_one, item_idx_two, similarity]`
        :return: cropped similarity matrix
        """
        return (
            similarity_matrix.withColumn(
                "similarity_order",
                sf.row_number().over(
                    Window.partitionBy("item_idx_one").orderBy(
                        sf.col("similarity").desc(),
                        sf.col("item_idx_two").desc(),
                    )
                ),
            )
            .filter(sf.col("similarity_order") <= self.num_neighbours)
            .drop("similarity_order")
        )

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        df = dataset.interactions.select(self.query_column, self.item_column, self.rating_column)
        if not self.use_rating:
            df = df.withColumn(self.rating_column, sf.lit(1))

        similarity_matrix = self._get_similarity(df)
        self.similarity = self._get_k_most_similar(similarity_matrix)
        self.similarity.cache().count()



import logging
from typing import Optional

import numpy as np

from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_inferers.hnswlib_filter_index_inferer import HnswlibFilterIndexInferer
from replay.models.extensions.ann.index_inferers.hnswlib_index_inferer import HnswlibIndexInferer
from replay.models.extensions.ann.utils import create_hnswlib_index_instance
from replay.utils import SparkDataFrame
from replay.utils.spark_utils import spark_to_pandas

from .base_index_builder import IndexBuilder

logger = logging.getLogger("replay")


class DriverHnswlibIndexBuilder(IndexBuilder):
    """
    Builder that builds hnswlib index on driver.
    """

    def produce_inferer(self, filter_seen_items: bool) -> IndexInferer:
        if filter_seen_items:
            return HnswlibFilterIndexInferer(self.index_params, self.index_store)
        else:
            return HnswlibIndexInferer(self.index_params, self.index_store)

    def build_index(
        self,
        vectors: SparkDataFrame,
        features_col: str,
        ids_col: Optional[str] = None,
    ):
        vectors = spark_to_pandas(vectors, self.allow_collect_to_master)
        vectors_np = np.squeeze(vectors[features_col].values)
        index = create_hnswlib_index_instance(self.index_params, init=True)

        if ids_col is not None:
            index.add_items(np.stack(vectors_np), vectors[ids_col].values)
        else:
            index.add_items(np.stack(vectors_np))

        self.index_store.save_to_store(lambda path: index.save_index(path))



from .entities.hnswlib_param import HnswlibParam
from .entities.nmslib_hnsw_param import NmslibHnswParam


def create_hnswlib_index_instance(params: HnswlibParam, init: bool = False):
    """
    Creates and returns hnswlib index

    :param params: `HnswlibParam`
    :param init: If `True` it will call the `init_index` method on the index.
        Used when we want to create a new index.
        If `False` then the index will be used to load index data from a file.
    :return: `hnswlib` index instance
    """
    import hnswlib

    index = hnswlib.Index(space=params.space, dim=params.dim)

    if init:
        # Initializing index - the maximum number of elements should be known beforehand
        index.init_index(
            max_elements=params.max_elements,
            ef_construction=params.ef_c,
            M=params.m,
        )

    return index


def create_nmslib_index_instance(params: NmslibHnswParam):
    """
    Creates and returns nmslib index

    :param params: `NmslibHnswParam`
    :return: `nmslib` index
    """
    import nmslib

    index = nmslib.init(
        method=params.method,
        space=params.space,
        data_type=nmslib.DataType.SPARSE_VECTOR,
    )

    return index



import warnings
from collections.abc import Generator, Sequence
from typing import TYPE_CHECKING, NamedTuple, Optional, Union, cast

import numpy as np
import torch
from torch.utils.data import Dataset as TorchDataset

if TYPE_CHECKING:
    from .schema import TensorFeatureInfo, TensorMap, TensorSchema
    from .sequential_dataset import SequentialDataset


# We do not use dataclasses as PyTorch default collate
# function in dataloader supports only namedtuple
class TorchSequentialBatch(NamedTuple):
    """
    Batch of TorchSequentialDataset
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: "TensorMap"


class TorchSequentialDataset(TorchDataset):
    """
    Torch dataset for sequential recommender models
    """

    def __init__(
        self,
        sequential: "SequentialDataset",
        max_sequence_length: int,
        sliding_window_step: Optional[int] = None,
        padding_value: Optional[int] = None,
    ) -> None:
        """
        :param sequential: sequential dataset
        :param max_sequence_length: the maximum length of sequence
        :param sliding_window_step: value of offset from each sequence start during iteration,
            `None` means the offset will be equals to difference between actual sequence
            length and `max_sequence_length`.
            Default: `None`
        :param padding_value: value to pad sequences to desired length
        """
        super().__init__()
        self._sequential = sequential
        self._max_sequence_length = max_sequence_length
        self._sliding_window_step = sliding_window_step
        if padding_value is not None:
            warnings.warn(
                "`padding_value` parameter will be removed in future versions. "
                "Instead, you should specify `padding_value` for each column in TensorSchema",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            padding_value = 0
        self._padding_value = padding_value
        self._index2sequence_map = self._build_index2sequence_map()

    def __len__(self) -> int:
        return len(self._index2sequence_map)

    def __getitem__(self, index: int) -> TorchSequentialBatch:
        sequence_index, sequence_offset = self._index2sequence_map[index]

        query_id = self._generate_query_id(sequence_index)
        padding_mask = self._generate_padding_mask(sequence_index)

        tensor_features = {
            feature_name: self._generate_tensor_feature(feature, sequence_index, sequence_offset)
            for feature_name, feature in self._sequential.schema.items()
        }

        return TorchSequentialBatch(
            query_id,
            padding_mask,
            tensor_features,
        )

    def _generate_query_id(self, sequence_index: int) -> torch.LongTensor:
        query_id = self._sequential.get_query_id(sequence_index)
        return torch.LongTensor([query_id])

    def _generate_padding_mask(self, sequence_index: int) -> torch.BoolTensor:
        mask = torch.ones(self._max_sequence_length, dtype=torch.bool)

        actual_sequence_len = self._sequential.get_sequence_length(sequence_index)
        if actual_sequence_len < self._max_sequence_length:
            mask[:-actual_sequence_len].fill_(0)

        return cast(torch.BoolTensor, mask)

    def _generate_tensor_feature(
        self,
        feature: "TensorFeatureInfo",
        sequence_index: int,
        sequence_offset: int,
    ) -> torch.Tensor:
        sequence = self._sequential.get_sequence(sequence_index, feature.name)
        if feature.is_seq:
            sequence = sequence[sequence_offset : sequence_offset + self._max_sequence_length]
        tensor_dtype = self._get_tensor_dtype(sequence)
        tensor_sequence = torch.tensor(sequence, dtype=tensor_dtype)
        if feature.is_seq:
            tensor_sequence = self._pad_sequence(tensor_sequence, feature.padding_value)

        return tensor_sequence

    def _pad_sequence(self, sequence: torch.Tensor, padding_value: int) -> torch.Tensor:
        assert len(sequence) <= self._max_sequence_length
        if len(sequence) == self._max_sequence_length:
            return sequence

        # form shape for padded_sequence. Now supported one and two-dimentions features
        padded_sequence_shape: Union[tuple[int, int], tuple[int]]
        if len(sequence.shape) == 1:
            padded_sequence_shape = (self._max_sequence_length,)
        elif len(sequence.shape) == 2:
            padded_sequence_shape = (self._max_sequence_length, sequence.shape[1])
        else:
            msg = f"Unsupported shape for sequence: {len(sequence.shape)}"
            raise ValueError(msg)

        padded_sequence = torch.full(
            padded_sequence_shape,
            padding_value,
            dtype=sequence.dtype,
        )
        padded_sequence[-len(sequence) :].copy_(sequence)
        return padded_sequence

    def _get_tensor_dtype(self, array: np.array) -> torch.dtype:
        if np.issubdtype(array.dtype, np.integer):
            return torch.long
        if np.issubdtype(array.dtype, np.floating):
            return torch.float32
        assert False, "Unknown tensor feature type"

    def _build_index2sequence_map(self) -> Sequence[tuple[int, int]]:
        return list(self._iter_with_window())

    def _iter_with_window(self) -> Generator[tuple[int, int], None, None]:
        for i in range(len(self._sequential)):
            actual_seq_len = self._sequential.get_sequence_length(i)
            left_seq_len = actual_seq_len - self._max_sequence_length

            if self._sliding_window_step is not None:
                offset_from_seq_beginning = left_seq_len
                while offset_from_seq_beginning > 0:
                    yield (i, offset_from_seq_beginning)
                    offset_from_seq_beginning -= self._sliding_window_step

                assert offset_from_seq_beginning <= 0
                yield (i, 0)
            else:
                offset_from_seq_beginning = max(0, left_seq_len)
                yield (i, offset_from_seq_beginning)


class TorchSequentialValidationBatch(NamedTuple):
    """
    Batch of TorchSequentialValidationDataset
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: "TensorMap"
    ground_truth: torch.LongTensor
    train: torch.LongTensor


DEFAULT_GROUND_TRUTH_PADDING_VALUE = -1
DEFAULT_TRAIN_PADDING_VALUE = -2


class TorchSequentialValidationDataset(TorchDataset):
    """
    Torch dataset for sequential recommender models that additionally stores ground truth
    """

    def __init__(
        self,
        sequential: "SequentialDataset",
        ground_truth: "SequentialDataset",
        train: "SequentialDataset",
        max_sequence_length: int,
        padding_value: Optional[int] = None,
        sliding_window_step: Optional[int] = None,
        label_feature_name: Optional[str] = None,
    ):
        """
        :param sequential: validation sequential dataset
        :param ground_truth: validation ground_truth sequential dataset
        :param train: train sequential dataset
        :param max_sequence_length: the maximum length of sequence
        :param padding_value: value to pad sequences to desired length
        :param sliding_window_step: value of offset from each sequence start during iteration,
            `None` means the offset will be equals to difference between actual sequence
            length and `max_sequence_length`.
            Default: `None`
        :param label_feature_name: the name of the column containing the sequence of items.
        """
        self._check_if_schema_match(sequential.schema, ground_truth.schema)
        self._check_if_schema_match(sequential.schema, train.schema)

        if label_feature_name:
            if label_feature_name not in ground_truth.schema:
                msg = "Label feature name not found in ground truth schema"
                raise ValueError(msg)

            if label_feature_name not in train.schema:
                msg = "Label feature name not found in train schema"
                raise ValueError(msg)

            if not ground_truth.schema[label_feature_name].is_cat:
                msg = "Label feature must be categorical"
                raise ValueError(msg)

            if not ground_truth.schema[label_feature_name].is_seq:
                msg = "Label feature must be sequential"
                raise ValueError(msg)

        if len(np.intersect1d(sequential.get_all_query_ids(), ground_truth.get_all_query_ids())) == 0:
            msg = "Sequential data and ground truth must contain the same query IDs"
            raise ValueError(msg)

        self._ground_truth = ground_truth
        self._train = train
        self._item_count = ground_truth.schema.item_id_features.item().cardinality
        self._label_feature_name = label_feature_name or ground_truth.schema.item_id_feature_name
        self._max_ground_truth_length = ground_truth.get_max_sequence_length()
        self._max_train_length = train.get_max_sequence_length()

        self._inner = TorchSequentialDataset(
            sequential=sequential,
            max_sequence_length=max_sequence_length,
            sliding_window_step=sliding_window_step,
            padding_value=padding_value,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> TorchSequentialValidationBatch:
        query_id, padding_mask, features = self._inner[index]

        return TorchSequentialValidationBatch(
            query_id=query_id,
            padding_mask=padding_mask,
            features=features,
            ground_truth=self._get_ground_truth(int(query_id.item())),
            train=self._get_train(int(query_id.item())),
        )

    def _get_ground_truth(self, query_id: int) -> torch.LongTensor:
        assert self._label_feature_name

        ground_truth_sequence = self._ground_truth.get_sequence_by_query_id(
            query_id,
            self._label_feature_name,
        )

        placeholder = np.full(self._max_ground_truth_length, DEFAULT_GROUND_TRUTH_PADDING_VALUE, dtype=np.int64)
        np.copyto(placeholder[: len(ground_truth_sequence)], ground_truth_sequence)
        return torch.LongTensor(placeholder)

    def _get_train(self, query_id: int) -> torch.LongTensor:
        assert self._label_feature_name

        train_sequence = self._train.get_sequence_by_query_id(
            query_id,
            self._label_feature_name,
        )

        placeholder = np.full(self._max_train_length, DEFAULT_TRAIN_PADDING_VALUE, dtype=np.int64)
        np.copyto(placeholder[: len(train_sequence)], train_sequence)
        return torch.LongTensor(placeholder)

    @classmethod
    def _check_if_schema_match(
        cls,
        sequential_schema: "TensorSchema",
        ground_truth_schema: "TensorSchema",
    ) -> None:
        sequential_item_feature = sequential_schema.item_id_features.item()
        ground_truth_item_feature = ground_truth_schema.item_id_features.item()

        if sequential_item_feature.name != ground_truth_item_feature.name:
            msg = "Schema mismatch: item feature name does not match ground truth"
            raise ValueError(msg)

        if sequential_item_feature.cardinality != ground_truth_item_feature.cardinality:
            msg = "Schema mismatch: item feature cardinality does not match ground truth"
            raise ValueError(msg)



import logging
import os
import shutil
import tempfile
import weakref
from typing import Any, Callable

from pyarrow import fs

from replay.utils import PYSPARK_AVAILABLE
from replay.utils.session_handler import State

from .base_index_store import IndexStore
from .utils import FileSystem, get_filesystem

if PYSPARK_AVAILABLE:
    from pyspark import SparkFiles


logger = logging.getLogger("replay")


if PYSPARK_AVAILABLE:

    class SparkFilesIndexStore(IndexStore):
        """Class that responsible for index store in spark files.
        Works through SparkContext.addFile()."""

        def _clean_up(self):  # pragma: no cover
            """Removes directory with index files
            before the instance is garbage collected."""
            if self.index_dir_path:
                shutil.rmtree(self.index_dir_path)

        def __init__(self, cleanup: bool = True):
            self.index_dir_path = None
            super().__init__(cleanup)
            if self.cleanup:
                weakref.finalize(self, self._clean_up)

        def load_index(
            self,
            init_index: Callable[[], None],
            load_index: Callable[[Any, str], None],
            configure_index: Callable[[Any], None],
        ):  # pragma: no cover
            if self._index:
                return self._index

            self._index = init_index()

            temp_file_path = SparkFiles.get("index")
            load_index(self._index, temp_file_path)
            configure_index(self._index)

            return self._index

        def save_to_store(self, save_index: Callable[[str], None]):
            self.index_dir_path: str = tempfile.mkdtemp()
            temp_file_path = os.path.join(self.index_dir_path, "index")
            save_index(temp_file_path)

            spark = State().session
            for filename in os.listdir(self.index_dir_path):
                index_file_path = os.path.join(self.index_dir_path, filename)
                spark.sparkContext.addFile("file://" + index_file_path)
                logger.info("Index file %s transferred to executors", index_file_path)

        def dump_index(self, target_path: str):
            destination_filesystem, target_path = fs.FileSystem.from_uri(target_path)
            target_path = os.path.join(target_path, "index_files")
            destination_filesystem.create_dir(target_path)
            fs.copy_files(
                self.index_dir_path,
                target_path,
                source_filesystem=fs.LocalFileSystem(),
                destination_filesystem=destination_filesystem,
            )

        def load_from_path(self, path: str):
            """Loads index from `path` directory to spark files."""
            path_info = get_filesystem(path)
            source_filesystem, path = fs.FileSystem.from_uri(
                path_info.hdfs_uri + path_info.path if path_info.filesystem == FileSystem.HDFS else path_info.path
            )
            path = os.path.join(path, "index_files")
            self.index_dir_path: str = tempfile.mkdtemp()
            fs.copy_files(
                path,
                self.index_dir_path,
                source_filesystem=source_filesystem,
                destination_filesystem=fs.LocalFileSystem(),
            )

            spark = State().session
            for filename in os.listdir(self.index_dir_path):
                index_file_path = os.path.join(self.index_dir_path, filename)
                spark.sparkContext.addFile("file://" + index_file_path)
                logger.info("Index file %s transferred to executors", index_file_path)



import abc
import json
import pickle
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Generic, Optional, TypeVar, Union

import numpy as np
import polars as pl
from pandas import DataFrame as PandasDataFrame
from polars import DataFrame as PolarsDataFrame

from replay.data import Dataset, FeatureHint, FeatureSchema, FeatureSource, FeatureType
from replay.data.dataset_utils import DatasetLabelEncoder
from replay.preprocessing import LabelEncoder, LabelEncodingRule
from replay.preprocessing.label_encoder import HandleUnknownStrategies

if TYPE_CHECKING:
    from .schema import TensorFeatureInfo, TensorFeatureSource, TensorSchema
    from .sequential_dataset import SequentialDataset

SequenceDataFrameLike = Union[PandasDataFrame, PolarsDataFrame]
_T = TypeVar("_T")


class SequenceTokenizer:
    """
    Data tokenizer for transformers;
    Encodes all categorical features (the ones marked as FeatureType.CATEGORICAL in
    the FeatureSchema) and stores all data as items sequences (sorted by time if a
    feature of type FeatureHint.TIMESTAMP is provided, unsorted otherwise).
    """

    def __init__(
        self,
        tensor_schema: "TensorSchema",
        handle_unknown_rule: HandleUnknownStrategies = "error",
        default_value_rule: Optional[Union[int, str]] = None,
        allow_collect_to_master: bool = False,
    ) -> None:
        """
        :param tensor_schema: tensor schema of tensor features
        :param handle_unknown_rule: handle unknown labels rule for LabelEncoder,
            values are in ('error', 'use_default_value', 'drop').
            Default: `error`
        :param default_value: Default value that will fill the unknown labels after transform.
            When the parameter handle_unknown is set to ``use_default_value``,
            this parameter is required and will set the encoded value of unknown labels.
            It has to be distinct from the values used to encode any of the labels in fit.
            If ``None``, then keep null.
            If ``int`` value, then fill by that value.
            If ``str`` value, should be \"last\" only, then fill by ``n_classes`` value.
            Default: ``None``.
        :param allow_collect_to_master: Flag allowing spark to make a collection to the master node,
            Default: ``False``.
        """
        self._tensor_schema = tensor_schema
        self._allow_collect_to_master = allow_collect_to_master
        self._encoder = DatasetLabelEncoder(
            handle_unknown_rule=handle_unknown_rule, default_value_rule=default_value_rule
        )
        self._check_tensor_schema(self._tensor_schema)

    def fit(self, dataset: Dataset) -> "SequenceTokenizer":
        """
        :param dataset: input dataset to fit

        :returns: fitted SequenceTokenizer
        """
        self._check_if_tensor_schema_matches_data(dataset, self._tensor_schema)
        self._assign_tensor_features_cardinality(dataset)
        self._encoder.fit(dataset)
        return self

    def transform(
        self,
        dataset: Dataset,
        tensor_features_to_keep: Optional[Sequence[str]] = None,
    ) -> "SequentialDataset":
        """
        :param dataset: input dataset to transform
        :param tensor_features_to_keep: specified feature names to transform
        :returns: SequentialDataset
        """
        self._check_if_tensor_schema_matches_data(dataset, self._tensor_schema, tensor_features_to_keep)
        return self._transform_unchecked(dataset, tensor_features_to_keep)

    def fit_transform(
        self,
        dataset: Dataset,
    ) -> "SequentialDataset":
        """
        :param dataset: input dataset to transform
        :returns: SequentialDataset
        """
        return self.fit(dataset)._transform_unchecked(dataset)

    @property
    def tensor_schema(self) -> "TensorSchema":
        """
        :returns: tensor schema
        """
        return self._tensor_schema

    @property
    def query_id_encoder(self) -> LabelEncoder:
        """
        :returns: encoder for query id
        """
        return self._encoder.query_id_encoder

    @property
    def item_id_encoder(self) -> LabelEncoder:
        """
        :returns: encoder for item id
        """
        return self._encoder.item_id_encoder

    @property
    def query_and_item_id_encoder(self) -> LabelEncoder:
        """
        :returns: encoder for query and item id
        """
        return self._encoder.query_and_item_id_encoder

    @property
    def interactions_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: encoder for interactions
        """
        return self._encoder.interactions_encoder

    @property
    def query_features_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: encoder for query features
        """
        return self._encoder.query_features_encoder

    @property
    def item_features_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: encoder for item features
        """
        return self._encoder.item_features_encoder

    def _transform_unchecked(
        self,
        dataset: Dataset,
        tensor_features_to_keep: Optional[Sequence[str]] = None,
    ) -> "SequentialDataset":
        from replay.data.nn.sequential_dataset import PandasSequentialDataset, PolarsSequentialDataset

        schema = self._tensor_schema
        if tensor_features_to_keep is not None:
            schema = schema.subset(tensor_features_to_keep)

        matched_dataset = self._match_features_with_tensor_schema(dataset, schema)

        encoded_dataset = self._encode_dataset(matched_dataset)
        is_polars = isinstance(encoded_dataset.interactions, PolarsDataFrame)
        grouped_interactions, query_features, item_features = self._group_dataset(encoded_dataset)
        sequence_features = self._make_sequence_features(
            schema,
            dataset.feature_schema,
            grouped_interactions,
            query_features,
            item_features,
        )

        assert self._tensor_schema.item_id_feature_name

        dataset_type = PolarsSequentialDataset if is_polars else PandasSequentialDataset

        return dataset_type(
            tensor_schema=schema,
            query_id_column=dataset.feature_schema.query_id_column,
            item_id_column=self._tensor_schema.item_id_feature_name,
            sequences=sequence_features,
        )

    def _encode_dataset(self, dataset: Dataset) -> Dataset:
        encoded_dataset = self._encoder.transform(dataset)
        return encoded_dataset

    def _group_dataset(
        self,
        dataset: Dataset,
    ) -> tuple[SequenceDataFrameLike, Optional[SequenceDataFrameLike], Optional[SequenceDataFrameLike]]:
        from replay.data.nn.utils import ensure_pandas, groupby_sequences

        grouped_interactions = groupby_sequences(
            events=dataset.interactions,
            groupby_col=dataset.feature_schema.query_id_column,
            sort_col=dataset.feature_schema.interactions_timestamp_column,
        )

        if isinstance(grouped_interactions, PolarsDataFrame):
            return (
                grouped_interactions.sort(dataset.feature_schema.query_id_column),
                dataset.query_features,
                dataset.item_features,
            )

        # We sort by QUERY_ID to make sure order is deterministic
        grouped_interactions_pd = ensure_pandas(
            grouped_interactions,
            self._allow_collect_to_master,
        )
        grouped_interactions_pd.sort_values(dataset.feature_schema.query_id_column, inplace=True, ignore_index=True)

        query_features_pd: Optional[PandasDataFrame] = None
        item_features_pd: Optional[PandasDataFrame] = None

        if dataset.query_features is not None:
            query_features_pd = ensure_pandas(dataset.query_features, self._allow_collect_to_master)
        if dataset.item_features is not None:
            item_features_pd = ensure_pandas(dataset.item_features, self._allow_collect_to_master)

        return grouped_interactions_pd, query_features_pd, item_features_pd

    def _make_sequence_features(
        self,
        schema: "TensorSchema",
        feature_schema: FeatureSchema,
        grouped_interactions: SequenceDataFrameLike,
        query_features: Optional[SequenceDataFrameLike],
        item_features: Optional[SequenceDataFrameLike],
    ) -> SequenceDataFrameLike:
        sequence_processor_class = (
            _PolarsSequenceProcessor if isinstance(grouped_interactions, PolarsDataFrame) else _PandasSequenceProcessor
        )

        processor = sequence_processor_class(
            tensor_schema=schema,
            query_id_column=feature_schema.query_id_column,
            item_id_column=feature_schema.item_id_column,
            grouped_interactions=grouped_interactions,
            query_features=query_features,
            item_features=item_features,
        )
        return processor.process_features()

    @classmethod
    def _match_features_with_tensor_schema(
        cls,
        dataset: Dataset,
        tensor_schema: "TensorSchema",
    ) -> Dataset:
        feature_subset_filter = cls._get_features_filter_from_schema(
            tensor_schema,
            query_id_column=dataset.feature_schema.query_id_column,
            item_id_column=dataset.feature_schema.item_id_column,
        )

        # We need to keep timestamp column in dataset as it's used to sort interactions
        timestamp_column = dataset.feature_schema.interactions_timestamp_column
        if timestamp_column:
            feature_subset_filter.add(timestamp_column)

        subset = dataset.subset(feature_subset_filter)
        return subset

    @classmethod
    def _get_features_filter_from_schema(
        cls,
        tensor_schema: "TensorSchema",
        query_id_column: str,
        item_id_column: str,
    ) -> set[str]:
        # We need only features, which related to tensor schema, otherwise feature should
        # be ignored for efficiency reasons. The code below does feature filtering, and
        # keeps features used as a source in tensor schema.

        # Query and item IDs are always needed
        features_subset: list[str] = [
            query_id_column,
            item_id_column,
        ]

        for tensor_feature in tensor_schema.values():
            for source in tensor_feature.feature_sources:
                assert source is not None

                # Some columns already added to encoder, skip them
                if source.column in features_subset:
                    continue

                if isinstance(source.source, FeatureSource):
                    features_subset.append(source.column)
                else:
                    assert False, "Unknown tensor feature source"

        return set(features_subset)

    @classmethod
    def _check_tensor_schema(cls, tensor_schema: "TensorSchema") -> None:
        # Check consistency of sequential features
        for tensor_feature in tensor_schema.all_features:
            feature_sources = tensor_feature.feature_sources
            if not feature_sources:
                msg = "All tensor features must have sources defined"
                raise ValueError(msg)

            source_tables: list[FeatureSource] = [s.source for s in feature_sources]

            unexpected_tables = list(filter(lambda x: not isinstance(x, FeatureSource), source_tables))
            if len(unexpected_tables) > 0:
                msg = f"Found unexpected source tables: {unexpected_tables}"
                raise ValueError(msg)

            if not tensor_feature.is_seq:
                if FeatureSource.INTERACTIONS in source_tables:
                    msg = "Interaction features must be treated as sequential"
                    raise ValueError(msg)

                if FeatureSource.ITEM_FEATURES in source_tables:
                    msg = "Item features must be treated as sequential"
                    raise ValueError(msg)

    @classmethod
    def _check_if_tensor_schema_matches_data(  # noqa: C901
        cls,
        dataset: Dataset,
        tensor_schema: "TensorSchema",
        tensor_features_to_keep: Optional[Sequence[str]] = None,
    ) -> None:
        # Check if all source columns specified in tensor schema exist in provided data frames
        sources_for_tensors: list["TensorFeatureSource"] = []
        for tensor_feature_name, tensor_feature in tensor_schema.items():
            if tensor_features_to_keep is not None and tensor_feature_name not in tensor_features_to_keep:
                continue

            if tensor_feature.feature_sources:
                sources_for_tensors += tensor_feature.feature_sources

        query_id_column = dataset.feature_schema.query_id_column
        item_id_column = dataset.feature_schema.item_id_column

        interaction_feature_columns = {
            *dataset.feature_schema.interaction_features.columns,
            query_id_column,
            item_id_column,
        }
        query_feature_columns = {*dataset.feature_schema.query_features.columns, query_id_column}
        item_feature_columns = {*dataset.feature_schema.item_features.columns, item_id_column}

        for feature_source in sources_for_tensors:
            assert feature_source is not None
            if feature_source.source == FeatureSource.INTERACTIONS:
                if feature_source.column not in interaction_feature_columns:
                    msg = f"Expected column '{feature_source.column}' in dataset"
                    raise ValueError(msg)
            elif feature_source.source == FeatureSource.QUERY_FEATURES:
                if dataset.query_features is None:
                    msg = f"Expected column '{feature_source.column}', but query features are not specified"
                    raise ValueError(msg)
                if feature_source.column not in query_feature_columns:
                    msg = f"Expected column '{feature_source.column}' in query features data frame"
                    raise ValueError(msg)
            elif feature_source.source == FeatureSource.ITEM_FEATURES:
                if dataset.item_features is None:
                    msg = f"Expected column '{feature_source.column}', but item features are not specified"
                    raise ValueError(msg)
                if feature_source.column not in item_feature_columns:
                    msg = f"Expected column '{feature_source.column}' in item features data frame"
                    raise ValueError(msg)
            else:
                msg = f"Found unexpected table '{feature_source.source}' in tensor schema"
                raise ValueError(msg)

        # Check if user ID and item ID columns are consistent with tensor schema
        if tensor_schema.query_id_feature_name is not None:
            tensor_feature = tensor_schema.query_id_features.item()
            assert tensor_feature.feature_source
            if tensor_feature.feature_source.column != dataset.feature_schema.query_id_column:
                msg = "Tensor schema query ID source colum does not match query ID in data frame"
                raise ValueError(msg)

        if tensor_schema.item_id_feature_name is None:
            msg = "Tensor schema must have item id feature defined"
            raise ValueError(msg)

        tensor_feature = tensor_schema.item_id_features.item()
        assert tensor_feature.feature_source
        if tensor_feature.feature_source.column != dataset.feature_schema.item_id_column:
            msg = "Tensor schema item ID source colum does not match item ID in data frame"
            raise ValueError(msg)

    def _assign_tensor_features_cardinality(self, dataset: Dataset) -> None:
        for tensor_feature in self._tensor_schema.categorical_features.all_features:
            dataset_feature = dataset.feature_schema[tensor_feature.feature_source.column]
            if tensor_feature.cardinality is not None:
                warnings.warn(
                    f"The specified cardinality of {tensor_feature.name} "
                    f"will be replaced by {dataset_feature.column} from Dataset"
                )
            if dataset_feature.feature_type not in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_LIST]:
                error_msg = (
                    f"TensorFeatureInfo {tensor_feature.name} "
                    f"and FeatureInfo {dataset_feature.column} must be the same FeatureType"
                )
                raise RuntimeError(error_msg)
            tensor_feature._set_cardinality(dataset_feature.cardinality)

    @classmethod
    def load(cls, path: str, use_pickle: bool = False, **kwargs) -> "SequenceTokenizer":
        """
        Load tokenizer object from the given path.

        :param path: Path to load the tokenizer.
        :param use_pickle: If `False` - tokenizer will be loaded from `.replay` directory.
            If `True` - tokenizer will be loaded with pickle.
            Default: `False`.

        :returns: Loaded tokenizer object.
        """
        from replay.data.nn import TensorSchema

        if not use_pickle:
            base_path = Path(path).with_suffix(".replay").resolve()
            with open(base_path / "init_args.json") as file:
                tokenizer_dict = json.loads(file.read())

            # load tensor_schema, tensor_features
            tensor_schema_data = tokenizer_dict["init_args"]["tensor_schema"]
            tokenizer_dict["init_args"]["tensor_schema"] = TensorSchema._create_object_by_args(tensor_schema_data)

            # Load encoder columns and rules
            types = list(FeatureHint) + list(FeatureSource)
            map_types = {x.name: x for x in types}
            encoder_features_columns = {
                map_types[key]: value for key, value in tokenizer_dict["encoder"]["features_columns"].items()
            }

            rules_dict = tokenizer_dict["encoder"]["encoding_rules"]
            for rule in rules_dict:
                rule_data = rules_dict[rule]
                if rule_data["mapping"] and rule_data["is_int"]:
                    rule_data["mapping"] = {int(key): value for key, value in rule_data["mapping"].items()}
                del rule_data["is_int"]

                tokenizer_dict["encoder"]["encoding_rules"][rule] = LabelEncodingRule(**rule_data)

            # Init tokenizer
            tokenizer = cls(**tokenizer_dict["init_args"])
            tokenizer._encoder._features_columns = encoder_features_columns
            tokenizer._encoder._encoding_rules = tokenizer_dict["encoder"]["encoding_rules"]
        else:
            warnings.warn(
                "with `use_pickle` equals to `True` will be deprecated in future versions",
                DeprecationWarning,
                stacklevel=2,
            )
            with open(path, "rb") as file:
                tokenizer = pickle.load(file)

        return tokenizer

    def save(self, path: str, use_pickle: bool = False) -> None:
        """
        Save the tokenizer to the given path.

        :param path: Path to save the tokenizer.
        :param use_pickle: If `False` - tokenizer will be saved in `.replay` directory.
            If `True` - tokenizer will be saved with pickle.
            Default: `False`.
        """
        if not use_pickle:
            tokenizer_dict = {}
            tokenizer_dict["_class_name"] = self.__class__.__name__
            tokenizer_dict["init_args"] = {
                "allow_collect_to_master": self._allow_collect_to_master,
                "handle_unknown_rule": self._encoder._handle_unknown_rule,
                "default_value_rule": self._encoder._default_value_rule,
                "tensor_schema": self._tensor_schema._get_object_args(),
            }

            # save DatasetLabelEncoder
            tokenizer_dict["encoder"] = {
                "features_columns": {key.name: value for key, value in self._encoder._features_columns.items()},
                "encoding_rules": {
                    key: {
                        "column": value.column,
                        "mapping": value._mapping,
                        "handle_unknown": value._handle_unknown,
                        "default_value": value._default_value,
                        "is_int": isinstance(next(iter(value._mapping.keys())), int),
                    }
                    for key, value in self._encoder._encoding_rules.items()
                },
            }

            base_path = Path(path).with_suffix(".replay").resolve()
            base_path.mkdir(parents=True, exist_ok=True)

            with open(base_path / "init_args.json", "w+") as file:
                json.dump(tokenizer_dict, file)
        else:
            warnings.warn(
                "with `use_pickle` equals to `True` will be deprecated in future versions",
                DeprecationWarning,
                stacklevel=2,
            )
            with open(path, "wb") as file:
                pickle.dump(self, file)


class _BaseSequenceProcessor(Generic[_T]):
    """
    Base class for sequence processing
    """

    def __init__(
        self,
        tensor_schema: "TensorSchema",
        query_id_column: str,
        item_id_column: str,
        grouped_interactions: _T,
        query_features: Optional[_T] = None,
        item_features: Optional[_T] = None,
    ) -> None:
        self._tensor_schema = tensor_schema
        self._query_id_column = query_id_column
        self._item_id_column = item_id_column
        self._grouped_interactions = grouped_interactions
        self._query_features = query_features
        self._item_features = item_features

    @abc.abstractmethod
    def process_features(self) -> _T:  # pragma: no cover
        """
        For each feature that you want to process, you should call the _process_feature function.
        """

    def _process_feature(self, tensor_feature_name: str) -> _T:
        """
        Process each tensor feature for dataframes.

        :param tensor_feature_name: name of feature to process.

        :returns: values for provided tensor_feature_name column
        """
        tensor_feature = self._tensor_schema[tensor_feature_name]
        if tensor_feature.is_cat:
            return self._process_cat_feature(tensor_feature)
        if tensor_feature.is_num:
            return self._process_num_feature(tensor_feature)
        assert False, "Unknown tensor feature type"

    def _process_num_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:
        """
        Process numerical tensor feature depends on it source.
        """
        assert tensor_feature.feature_sources is not None
        if tensor_feature.feature_source.source == FeatureSource.INTERACTIONS:
            return self._process_num_interaction_feature(tensor_feature)
        if tensor_feature.feature_source.source == FeatureSource.QUERY_FEATURES:
            return self._process_num_query_feature(tensor_feature)
        if tensor_feature.feature_source.source == FeatureSource.ITEM_FEATURES:
            return self._process_num_item_feature(tensor_feature)
        assert False, "Unknown tensor feature source table"

    def _process_cat_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:
        """
        Process categorical tensor feature depends on it source.
        """
        assert tensor_feature.feature_sources is not None
        if tensor_feature.feature_source.source == FeatureSource.INTERACTIONS:
            return self._process_cat_interaction_feature(tensor_feature)
        if tensor_feature.feature_source.source == FeatureSource.QUERY_FEATURES:
            return self._process_cat_query_feature(tensor_feature)
        if tensor_feature.feature_source.source == FeatureSource.ITEM_FEATURES:
            return self._process_cat_item_feature(tensor_feature)
        assert False, "Unknown tensor feature source table"

    @abc.abstractmethod
    def _process_cat_interaction_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass

    @abc.abstractmethod
    def _process_cat_query_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass

    @abc.abstractmethod
    def _process_cat_item_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass

    @abc.abstractmethod
    def _process_num_interaction_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass

    @abc.abstractmethod
    def _process_num_query_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass

    @abc.abstractmethod
    def _process_num_item_feature(self, tensor_feature: "TensorFeatureInfo") -> _T:  # pragma: no cover
        pass


class _PandasSequenceProcessor(_BaseSequenceProcessor[PandasDataFrame]):
    """
    Class to process sequences of different categorical and numerical features.

    Processing performs over all features in `tensor_schema`. Each feature processing steps
    depends on feature type (categorical/numerical), feature source (interactions/query features/item features)
    and `grouped_interactions` in Pandas DataFrame format.
    """

    def __init__(
        self,
        tensor_schema: "TensorSchema",
        query_id_column: str,
        item_id_column: str,
        grouped_interactions: PandasDataFrame,
        query_features: Optional[PandasDataFrame] = None,
        item_features: Optional[PandasDataFrame] = None,
    ) -> None:
        super().__init__(
            tensor_schema=tensor_schema,
            query_id_column=query_id_column,
            item_id_column=item_id_column,
            grouped_interactions=grouped_interactions,
            query_features=(
                query_features.set_index(query_id_column).sort_index() if query_features is not None else None
            ),
            item_features=item_features.set_index(item_id_column).sort_index() if item_features is not None else None,
        )

    def process_features(self) -> PandasDataFrame:
        """
        :returns: processed Pandas DataFrame with all features from tensor schema.
        """
        all_features: dict[str, Union[np.ndarray, list[np.ndarray]]] = {}
        all_features[self._query_id_column] = self._grouped_interactions[self._query_id_column].values

        for tensor_feature_name in self._tensor_schema:
            all_features[tensor_feature_name] = self._process_feature(tensor_feature_name)

        return PandasDataFrame(all_features)

    def _process_num_interaction_feature(
        self, tensor_feature: "TensorFeatureInfo"
    ) -> Union[list[np.ndarray], list[list]]:
        """
        Process numerical interaction feature.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        assert tensor_feature.is_seq

        source = tensor_feature.feature_source
        assert source is not None

        values = []
        for sequence in self._grouped_interactions[source.column].values:
            if tensor_feature.feature_type == FeatureType.NUMERICAL_LIST:
                values.append(list(sequence))
            else:
                values.append(np.array(sequence))
        return values

    def _process_num_item_feature(self, tensor_feature: "TensorFeatureInfo") -> Union[list[np.ndarray], list[list]]:
        """
        Process numerical feature from item features dataset.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        assert tensor_feature.is_seq
        assert self._item_features is not None

        source = tensor_feature.feature_source
        assert source is not None

        item_feature = self._item_features[source.column]
        values = []

        for item_id_sequence in self._grouped_interactions[self._item_id_column]:
            feature_sequence = item_feature.loc[item_id_sequence].values
            if tensor_feature.feature_type == FeatureType.NUMERICAL_LIST:
                values.append(feature_sequence.tolist())
            else:
                values.append(np.array(feature_sequence))

        return values

    def _process_num_query_feature(self, tensor_feature: "TensorFeatureInfo") -> list[np.ndarray]:
        """
        Process numerical feature from query features dataset.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        return self._process_cat_query_feature(tensor_feature)

    def _process_cat_interaction_feature(
        self, tensor_feature: "TensorFeatureInfo"
    ) -> Union[list[np.ndarray], list[list]]:
        """
        Process categorical interaction feature.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        assert tensor_feature.is_seq

        source = tensor_feature.feature_source
        assert source is not None

        values = []
        for sequence in self._grouped_interactions[source.column].values:
            if tensor_feature.feature_type == FeatureType.CATEGORICAL_LIST:
                values.append(list(sequence))
            else:
                values.append(np.array(sequence))
        return values

    def _process_cat_query_feature(self, tensor_feature: "TensorFeatureInfo") -> list[np.ndarray]:
        """
        Process categorical feature from query features dataset.

        :param tensor_feature: tensor feature information.

        :returns: sequences with length of item sequence for each query for
            sequential features and one size sequences otherwise.
        """
        assert self._query_features is not None

        source = tensor_feature.feature_source
        assert source is not None

        query_feature = self._query_features[source.column].values
        if tensor_feature.is_seq:
            if tensor_feature.is_list:
                result = []
                for i, item_id_sequence in enumerate(self._grouped_interactions[self._item_id_column]):
                    seq_len = len(item_id_sequence)
                    result.append(np.repeat(query_feature[i], seq_len).reshape(-1, seq_len).T)
                return result
            else:
                return [
                    np.full(len(item_id_sequence), query_feature[i])
                    for i, item_id_sequence in enumerate(self._grouped_interactions[self._item_id_column])
                ]
        return [np.array([query_feature[i]]).reshape(-1) for i in range(len(self._grouped_interactions))]

    def _process_cat_item_feature(self, tensor_feature: "TensorFeatureInfo") -> Union[list[np.ndarray], list[list]]:
        """
        Process categorical feature from item features dataset.

        :param tensor_feature: tensor feature information.

        :returns: item features as a sequence for each item in a sequence
            for each query.
        """
        assert tensor_feature.is_seq
        assert self._item_features is not None

        source = tensor_feature.feature_source
        assert source is not None

        item_feature = self._item_features[source.column]
        values: list[np.ndarray] = []

        for item_id_sequence in self._grouped_interactions[self._item_id_column]:
            feature_sequence = item_feature.loc[item_id_sequence].values
            if tensor_feature.feature_type == FeatureType.CATEGORICAL_LIST:
                values.append(feature_sequence.tolist())
            else:
                values.append(np.array(feature_sequence, dtype=np.int64))

        return values


class _PolarsSequenceProcessor(_BaseSequenceProcessor[PolarsDataFrame]):
    """
    Class to process sequences of different categorical and numerical features.

    Processing performs over all features in `tensor_schema`. Each feature processing steps
    depends on feature type (categorical/numerical), feature source (interactions/query features/item features)
    and `grouped_interactions` in Polars DataFrame format.
    """

    def process_features(self) -> PolarsDataFrame:
        """
        :returns: processed Polars DataFrame with all features from tensor schema.
        """
        data = self._grouped_interactions.select(self._query_id_column)
        for tensor_feature_name in self._tensor_schema:
            data = data.join(self._process_feature(tensor_feature_name), on=self._query_id_column, how="left")
        return data

    def _process_num_interaction_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process numerical interaction feature.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        return self._process_cat_interaction_feature(tensor_feature)

    def _process_num_query_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process numerical feature from query features dataset.

        :param tensor_feature: tensor feature information.

        :returns: sequences with length of item sequence for each query for
            sequential features and one size sequences otherwise.
        """
        return self._process_cat_query_feature(tensor_feature)

    def _process_num_item_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process numerical feature from item features dataset.

        :param tensor_feature: tensor feature information.

        :returns: item features as a sequence for each item in a sequence
            for each query.
        """
        return self._process_cat_item_feature(tensor_feature)

    def _process_cat_interaction_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process categorical interaction feature.

        :param tensor_feature: tensor feature information.

        :returns: tensor feature column as a sequences from `grouped_interactions`.
        """
        assert tensor_feature.is_seq

        source = tensor_feature.feature_source
        assert source is not None

        return self._grouped_interactions.select(self._query_id_column, source.column).rename(
            {source.column: tensor_feature.name}
        )

    def _process_cat_query_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process categorical feature from query features dataset.

        :param tensor_feature: tensor feature information.

        :returns: sequences with length of item sequence for each query for
            sequential features and one size sequences otherwise.
        """
        assert self._query_features is not None

        source = tensor_feature.feature_source
        assert source is not None

        if not tensor_feature.is_seq:
            result = self._query_features.select(self._query_id_column, source.column).rename(
                {source.column: tensor_feature.name}
            )
            if not tensor_feature.is_list:
                result = result.with_columns(pl.col(tensor_feature.name).cast(pl.List(pl.Int64)))
            return result

        lengths = self._grouped_interactions.select(
            self._query_id_column, pl.col(self._item_id_column).list.len().alias("len")
        )
        result = lengths.join(
            self._query_features.select(self._query_id_column, source.column), on=self._query_id_column, how="left"
        )

        if tensor_feature.is_list:
            return (
                result.map_rows(
                    lambda x: (
                        x[0],
                        [x[2]] * x[1],
                    )
                )
            ).rename({"column_0": self._query_id_column, "column_1": tensor_feature.name})

        # just categorical branch
        return result.select(self._query_id_column, pl.col(source.column).repeat_by("len")).rename(
            {source.column: tensor_feature.name}
        )

    def _process_cat_item_feature(self, tensor_feature: "TensorFeatureInfo") -> PolarsDataFrame:
        """
        Process categorical feature from item features dataset.

        :param tensor_feature: tensor feature information.

        :returns: item features as a sequence for each item in a sequence
            for each query.
        """
        assert tensor_feature.is_seq
        assert self._item_features is not None

        source = tensor_feature.feature_source
        assert source is not None
        return (
            self._grouped_interactions.select(self._query_id_column, self._item_id_column)
            .map_rows(
                lambda x: (
                    x[0],
                    self._item_features.select(source.column)
                    .filter(self._item_features[self._item_id_column].is_in(x[1]))
                    .to_series()
                    .to_list(),
                )
            )
            .rename({"column_0": self._query_id_column, "column_1": tensor_feature.name})
        )



from abc import ABC, abstractmethod
from typing import Optional

from replay.models.extensions.ann.entities.base_hnsw_param import BaseHnswParam
from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_stores.base_index_store import IndexStore
from replay.utils import SparkDataFrame


class IndexBuilder(ABC):
    """Abstract base class for index builders.
    Describes a common interface for index builders. And provides common methods for them.
    """

    def __init__(
        self,
        index_params: BaseHnswParam,
        index_store: IndexStore,
        allow_collect_to_master: bool = False,
    ):
        self.index_store = index_store
        self.index_params = index_params
        self.allow_collect_to_master = allow_collect_to_master

    @abstractmethod
    def build_index(
        self,
        vectors: SparkDataFrame,
        features_col: str,
        ids_col: Optional[str] = None,
    ):
        """
        Method that builds index and stores it using the `IndexStore` class.

        :param vectors: SparkDataFrame with vectors to build index.
            Schema: [{ids_col}: int, {features_col}: array<float>]
            or [{features_col}: array<float>]
        :param features_col: Name of column from `vectors` dataframe
            that contains vectors to build index
        :param ids_col: Name of column that contains identifiers of vectors.
            None if `vectors` dataframe have no id column.
        """

    @abstractmethod
    def produce_inferer(self, filter_seen_items: bool) -> IndexInferer:
        """Method that produce `IndexInferer`."""

    def init_meta_as_dict(self):
        """
        Returns meta-information for class instance initialization.
        Used to save the index builder to disk.
        :return: dictionary with init meta.
        """
        return {
            "builder": {
                "module": type(self).__module__,
                "class": type(self).__name__,
            },
            "index_param": self.index_params.init_meta_as_dict(),
        }



import logging
import os
import tempfile
import weakref
from typing import Any, Callable

from pyarrow import fs

from .base_index_store import IndexStore
from .utils import FileSystem, get_filesystem

logger = logging.getLogger("replay")


class HdfsIndexStore(IndexStore):
    """Class that responsible for index store in HDFS."""

    def __init__(self, warehouse_dir: str, index_dir: str):
        index_dir_path = os.path.join(warehouse_dir, index_dir)
        self._index_dir_info = get_filesystem(index_dir_path)
        if self._index_dir_info.filesystem != FileSystem.HDFS:
            msg = f"Can't recognize path {index_dir_path} as HDFS path!"
            raise ValueError(msg)
        self._hadoop_fs = fs.HadoopFileSystem.from_uri(self._index_dir_info.hdfs_uri)
        super().__init__()

        if self.cleanup:
            logger.debug(
                "Index directory %s is marked for deletion via weakref.finalize()",
                self._index_dir_info.path,
            )
            weakref.finalize(self, self._hadoop_fs.delete_dir, self._index_dir_info.path)

    def load_index(
        self,
        init_index: Callable[[], None],
        load_index: Callable[[Any, str], None],
        configure_index: Callable[[Any], None],
    ):
        if self._index:
            return self._index

        self._index = init_index()

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file_path = os.path.join(temp_dir, "index")

            # here we copy index files from hdfs directory
            # to local disk directory
            fs.copy_files(
                self._index_dir_info.path,
                "file://" + temp_dir,
                source_filesystem=self._hadoop_fs,
            )
            load_index(self._index, temp_file_path)
            configure_index(self._index)

        return self._index

    def save_to_store(self, save_index: Callable[[str], None]):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file_path = os.path.join(temp_dir, "index")
            save_index(temp_file_path)

            # here we copy index files from local disk directory
            # to hdfs directory
            fs.copy_files(
                "file://" + temp_dir,
                self._index_dir_info.path,
                destination_filesystem=self._hadoop_fs,
            )
            # param use_threads=True (?)
            logger.info("Index files saved to %s", self._index_dir_info.path)

    def dump_index(self, target_path: str):
        target_path_info = get_filesystem(target_path)
        destination_filesystem, target_path = fs.FileSystem.from_uri(
            target_path_info.hdfs_uri + target_path_info.path
            if target_path_info.filesystem == FileSystem.HDFS
            else target_path_info.path
        )
        target_path = os.path.join(target_path, "index_files")
        fs.copy_files(
            self._index_dir_info.path,
            target_path,
            source_filesystem=self._hadoop_fs,
            destination_filesystem=destination_filesystem,
        )
